#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.core import Node, Relationship, Path, PropertySet, LabelSet
from py2neo.batch.core import Batch, Job, CypherJob, Target
from py2neo.batch.push import PushNodeLabelsJob, PushPropertiesJob, PushPropertyJob


__all__ = ["CreateNodeJob", "CreateRelationshipJob", "CreatePathJob", "CreateUniquePathJob",
           "DeleteEntityJob", "DeletePropertyJob", "DeletePropertiesJob",
           "AddNodeLabelsJob", "RemoveNodeLabelJob", "WriteBatch"]


def _create_query(p, unique=False):
    nodes, path, values, params = [], [], [], {}

    def append_node(i, node):
        if node is None:
            path.append("(n{0})".format(i))
            values.append("n{0}".format(i))
        elif node.bound:
            path.append("(n{0})".format(i))
            nodes.append("n{0}=node({{i{0}}})".format(i))
            params["i{0}".format(i)] = node._id
            values.append("n{0}".format(i))
        else:
            path.append("(n{0} {{p{0}}})".format(i))
            params["p{0}".format(i)] = node.properties
            values.append("n{0}".format(i))

    def append_rel(i, rel):
        if rel.properties:
            path.append("-[r{0}:`{1}` {{q{0}}}]->".format(i, rel.type))
            params["q{0}".format(i)] = PropertySet(rel.properties)
            values.append("r{0}".format(i))
        else:
            path.append("-[r{0}:`{1}`]->".format(i, rel.type))
            values.append("r{0}".format(i))

    append_node(0, p.nodes[0])
    for i, rel in enumerate(p.rels):
        append_rel(i, rel)
        append_node(i + 1, p.nodes[i + 1])
    clauses = []
    if nodes:
        clauses.append("START {0}".format(",".join(nodes)))
    if unique:
        clauses.append("CREATE UNIQUE p={0}".format("".join(path)))
    else:
        clauses.append("CREATE p={0}".format("".join(path)))
    #clauses.append("RETURN {0}".format(",".join(values)))
    clauses.append("RETURN p")
    query = " ".join(clauses)
    return query, params


class CreateNodeJob(Job):

    target = Target("node")

    def __init__(self, **properties):
        Job.__init__(self, "POST", self.target, properties)


class CreateRelationshipJob(Job):

    def __init__(self, start_node, rel, end_node, **properties):
        body = {"type": rel.type, "to": Target(end_node).uri_string}
        if rel.properties or properties:
            body["data"] = dict(rel.properties, **properties)
        Job.__init__(self, "POST", Target(start_node, "relationships"), body)


class CreatePathJob(CypherJob):

    def __init__(self, *entities):
        CypherJob.__init__(self, *_create_query(Path(*entities)))


class CreateUniquePathJob(CypherJob):

    def __init__(self, *entities):
        CypherJob.__init__(self, *_create_query(Path(*entities), unique=True))


class DeleteEntityJob(Job):

    def __init__(self, entity):
        Job.__init__(self, "DELETE", Target(entity))


class DeletePropertyJob(Job):

    def __init__(self, entity, key):
        Job.__init__(self, "DELETE", Target(entity, "properties", key))


class DeletePropertiesJob(Job):

    def __init__(self, entity):
        Job.__init__(self, "DELETE", Target(entity, "properties"))


class AddNodeLabelsJob(Job):

    def __init__(self, node, *labels):
        Job.__init__(self, "POST", Target(node, "labels"), list(LabelSet(labels)))


class RemoveNodeLabelJob(Job):

    def __init__(self, entity, label):
        Job.__init__(self, "DELETE", Target(entity, "labels", label))


class WriteBatch(Batch):
    """ Generic batch execution facility for data write requests. Most methods
    return a :py:class:`BatchRequest <py2neo.neo4j.BatchRequest>` object that
    can be used as a reference in other methods. See the
    :py:meth:`create <py2neo.neo4j.WriteBatch.create>` method for an example
    of this.
    """

    def __init__(self, graph):
        Batch.__init__(self, graph)

    def run(self):
        self.graph.batch.run(self)

    def stream(self):
        for result in self.graph.batch.stream(self):
            yield result.content

    def submit(self):
        return [result.content for result in self.graph.batch.submit(self)]

    def create(self, abstract):
        """ Create a node or relationship based on the abstract entity
        provided. For example::

            batch = WriteBatch(graph)
            a = batch.create(node(name="Alice"))
            b = batch.create(node(name="Bob"))
            batch.create(rel(a, "KNOWS", b))
            results = batch.submit()

        :param abstract: node or relationship
        :type abstract: abstract
        :return: batch request object
        """
        entity = self.graph.cast(abstract)
        if isinstance(entity, Node):
            return self.append(CreateNodeJob(**entity.properties))
        elif isinstance(entity, Relationship):
            start_node = self.resolve(entity.start_node)
            end_node = self.resolve(entity.end_node)
            return self.append(CreateRelationshipJob(start_node, entity.rel, end_node))
        else:
            raise TypeError(entity)

    def create_path(self, node, *rels_and_nodes):
        """ Construct a path across a specified set of nodes and relationships.
        Nodes may be existing concrete node instances, abstract nodes or
        :py:const:`None` but references to other requests are not supported.

        :param node: start node
        :type node: concrete, abstract or :py:const:`None`
        :param rels_and_nodes: alternating relationships and nodes
        :type rels_and_nodes: concrete, abstract or :py:const:`None`
        :return: batch request object
        """
        return self.append(CreatePathJob(node, *rels_and_nodes))

    def get_or_create_path(self, node, *rels_and_nodes):
        """ Construct a unique path across a specified set of nodes and
        relationships, adding only parts that are missing. Nodes may be
        existing concrete node instances, abstract nodes or :py:const:`None`
        but references to other requests are not supported.

        :param node: start node
        :type node: concrete, abstract or :py:const:`None`
        :param rels_and_nodes: alternating relationships and nodes
        :type rels_and_nodes: concrete, abstract or :py:const:`None`
        :return: batch request object
        """
        return self.append(CreateUniquePathJob(node, *rels_and_nodes))

    def delete(self, entity):
        """ Delete a node or relationship from the graph.

        :param entity: node or relationship to delete
        :type entity: concrete or reference
        :return: batch request object
        """
        return self.append(DeleteEntityJob(self.resolve(entity)))

    def set_property(self, entity, key, value):
        """ Set a single property on a node or relationship.

        :param entity: node or relationship on which to set property
        :type entity: concrete or reference
        :param key: property key
        :type key: :py:class:`str`
        :param value: property value
        :return: batch request object
        """
        return self.append(PushPropertyJob(self.resolve(entity), key, value))

    def set_properties(self, entity, properties):
        """ Replace all properties on a node or relationship.

        :param entity: node or relationship on which to set properties
        :type entity: concrete or reference
        :param properties: properties
        :type properties: :py:class:`dict`
        :return: batch request object
        """
        return self.append(PushPropertiesJob(self.resolve(entity), properties))

    def delete_property(self, entity, key):
        """ Delete a single property from a node or relationship.

        :param entity: node or relationship from which to delete property
        :type entity: concrete or reference
        :param key: property key
        :type key: :py:class:`str`
        :return: batch request object
        """
        return self.append(DeletePropertyJob(self.resolve(entity), key))

    def delete_properties(self, entity):
        """ Delete all properties from a node or relationship.

        :param entity: node or relationship from which to delete properties
        :type entity: concrete or reference
        :return: batch request object
        """
        return self.append(DeletePropertiesJob(self.resolve(entity)))

    def add_labels(self, node, *labels):
        """ Add labels to a node.

        :param node: node to which to add labels
        :type entity: concrete or reference
        :param labels: text labels
        :type labels: :py:class:`str`
        :return: batch request object
        """
        return self.append(AddNodeLabelsJob(self.resolve(node), *labels))

    def remove_label(self, node, label):
        """ Remove a label from a node.

        :param node: node from which to remove labels (can be a reference to
            another request within the same batch)
        :param label: text label
        :type label: :py:class:`str`
        :return: batch request object
        """
        return self.append(RemoveNodeLabelJob(self.resolve(node), label))

    def set_labels(self, node, *labels):
        """ Replace all labels on a node.

        :param node: node on which to replace labels (can be a reference to
            another request within the same batch)
        :param labels: text labels
        :type labels: :py:class:`str`
        :return: batch request object
        """
        return self.append(PushNodeLabelsJob(self.resolve(node), labels))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json

from py2neo.core import Resource, UnmanagedExtension
from py2neo.util import version_tuple

from py2neo.ext.geoff.xmlutil import xml_to_geoff


__all__ = ["GeoffLoader", "NodeDictionary"]


class GeoffLoader(UnmanagedExtension):
    """ A wrapper for the `load2neo <http://nigelsmall.com/load2neo>`_
    unmanaged extension.
    """

    DEFAULT_PATH = "/load2neo/"

    def __init__(self, graph, path=DEFAULT_PATH):
        UnmanagedExtension.__init__(self, graph, path)
        self.geoff_loader = Resource(self.resource.metadata["geoff_loader"])

    @property
    def load2neo_version(self):
        """ The version of server extension currently installed.
        """
        return version_tuple(self.resource.metadata["load2neo_version"])

    def load(self, string):
        """ Load Geoff data from a string.
        """
        rs = self.geoff_loader.post(string)
        return [NodeDictionary(self.graph, json.loads(line))
                for line in rs.content.splitlines(False)]

    def load_xml(self, string):
        """ Load XML data by first converting to Geoff format.
        """
        return self.load(xml_to_geoff(string))


class NodeDictionary(object):
    """ A set of nodes returned from a successful Geoff load operation.
    Each node is keyed by the identifier used in the original Geoff
    data.
    """

    def __init__(self, graph, data):
        self.graph = graph
        self.__data = data

    def __len__(self):
        return len(self.__data)

    def __iter__(self):
        return iter(self.__data)

    def __getitem__(self, key):
        return self.graph.node(self.__data[key])

    def get_ref(self, key):
        """ Get the node reference string for the node identified by
        the given key.
        """
        return "node/%s" % self.__data[key]

    def keys(self):
        """ Fetch a list of all node keys.
        """
        return self.__data.keys()

    def values(self):
        """ Fetch a list of all nodes.
        """
        return self.__data.values()

    def items(self):
        """ Fetch a list of all key-node pairs, each returned as a 2-tuple.
        """
        return self.__data.items()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.core import LabelSet, Node, Path, PropertySet, Rel
from py2neo.batch.core import Batch, Job, Target


__all__ = ["PushPropertyJob", "PushPropertiesJob", "PushNodeLabelsJob", "PushBatch"]


class PushPropertyJob(Job):

    def __init__(self, entity, key, value):
        Job.__init__(self, "PUT", Target(entity, "properties", key), value)


class PushPropertiesJob(Job):

    def __init__(self, entity, properties):
        Job.__init__(self, "PUT", Target(entity, "properties"), PropertySet(properties))


class PushNodeLabelsJob(Job):

    def __init__(self, node, labels):
        Job.__init__(self, "PUT", Target(node, "labels"), list(LabelSet(labels)))


class PushBatch(Batch):
    """ A batch of push jobs.
    """

    def __init__(self, graph):
        Batch.__init__(self, graph)

    def append(self, entity):
        """ Append an entity to the list to be pushed.

        :param entity: An entity such as a :class:`py2neo.Node`.

        """
        if isinstance(entity, Node):
            self.jobs.append(PushPropertiesJob(entity, entity.properties))
            if entity.graph.supports_node_labels:
                self.jobs.append(PushNodeLabelsJob(entity, entity.labels))
        elif isinstance(entity, Rel):
            self.jobs.append(PushPropertiesJob(entity, entity.properties))
        elif isinstance(entity, Path):
            for relationship in entity.relationships:
                self.jobs.append(PushPropertiesJob(relationship, relationship.properties))
        else:
            raise TypeError("Cannot pull object of type %s" % entity.__class__.__name__)

    def push(self):
        """ Push details from all entities in this batch to their
        remote counterparts.
        """
        self.graph.batch.run(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__author__ = "Nigel Small <nigel@py2neo.org>"
__copyright__ = "2011-2014, Nigel Small"
__email__ = "nigel@py2neo.org"
__license__ = "Apache License, Version 2.0"
__package__ = "py2neo"
__version__ = "2.0"


from py2neo.core import *
from py2neo.error import *
from py2neo.legacy import LegacyNode
from py2neo.packages.httpstream.watch import watch


__all__ = ["Graph", "Node", "Relationship", "Path", "NodePointer", "Rel", "Rev", "Subgraph",
           "ServiceRoot", "PropertySet", "LabelSet", "PropertyContainer",
           "authenticate", "rewrite", "watch",
           "BindError", "Finished", "GraphError", "JoinError",
           "ServerPlugin", "UnmanagedExtension", "Service", "Resource", "ResourceTemplate",
           "LegacyNode", "node", "rel",
]


node = Node.cast
rel = Relationship.cast



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ["BindError", "Finished", "GraphError", "JoinError"]


class BindError(Exception):
    """ Raised when a local graph entity is not or cannot be bound
    to a remote graph entity.
    """


class Finished(Exception):
    """ Raised when actions are attempted against a finished object
    that is no longer available for use.
    """

    def __init__(self, obj):
        self.obj = obj

    def __repr__(self):
        return "%s finished" % self.obj.__class__.__name__


class GraphError(Exception):
    """ Default exception class for all errors returned by the
    Neo4j server.
    """

    __cause__ = None
    exception = None
    fullname = None
    request = None
    response = None
    stacktrace = None

    def __new__(cls, *args, **kwargs):
        try:
            exception = kwargs["exception"]
            try:
                error_cls = type(exception, (cls,), {})
            except TypeError:
                # for Python 2.x
                error_cls = type(str(exception), (cls,), {})
        except KeyError:
            error_cls = cls
        return Exception.__new__(error_cls, *args)

    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args)
        for key, value in kwargs.items():
            setattr(self, key, value)


class JoinError(Exception):
    """ Raised when two graph entities cannot be joined together.
    """



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import unicode_literals

import logging
import sys
import threading


__all__ = ["Watcher", "watch"]


def black(s):
    return "\x1b[30m{}\x1b[0m".format(s)


def red(s):
    return "\x1b[31m{}\x1b[0m".format(s)


def green(s):
    return "\x1b[32m{}\x1b[0m".format(s)


def yellow(s):
    return "\x1b[33m{}\x1b[0m".format(s)


def blue(s):
    return "\x1b[34m{}\x1b[0m".format(s)


def magenta(s):
    return "\x1b[35m{}\x1b[0m".format(s)


def cyan(s):
    return "\x1b[36m{}\x1b[0m".format(s)


def white(s):
    return "\x1b[36m{}\x1b[0m".format(s)


def bright_black(s):
    return "\x1b[30;1m{}\x1b[0m".format(s)


def bright_red(s):
    return "\x1b[31;1m{}\x1b[0m".format(s)


def bright_green(s):
    return "\x1b[32;1m{}\x1b[0m".format(s)


def bright_yellow(s):
    return "\x1b[33;1m{}\x1b[0m".format(s)


def bright_blue(s):
    return "\x1b[34;1m{}\x1b[0m".format(s)


def bright_magenta(s):
    return "\x1b[35;1m{}\x1b[0m".format(s)


def bright_cyan(s):
    return "\x1b[36;1m{}\x1b[0m".format(s)


def bright_white(s):
    return "\x1b[37;1m{}\x1b[0m".format(s)


class ColourFormatter(logging.Formatter):

    def format(self, record):
        s = super(ColourFormatter, self).format(record)
        if record.levelno == logging.CRITICAL:
            return bright_red(s)
        elif record.levelno == logging.ERROR:
            return bright_yellow(s)
        elif record.levelno == logging.WARNING:
            return yellow(s)
        elif record.levelno == logging.INFO:
            return cyan(s)
        elif record.levelno == logging.DEBUG:
            return blue(s)
        else:
            return s


class Watcher(threading.local):

    handlers = {}

    def __init__(self, logger_name):
        super(Watcher, self).__init__()
        self.logger_name = logger_name
        self.logger = logging.getLogger(self.logger_name)
        self.formatter = ColourFormatter()

    def watch(self, level=None, out=sys.stdout):
        try:
            self.logger.removeHandler(self.handlers[self.logger_name])
        except KeyError:
            pass
        handler = logging.StreamHandler(out)
        handler.setFormatter(self.formatter)
        self.handlers[self.logger_name] = handler
        self.logger.addHandler(handler)
        if level is None:
            level = logging.DEBUG if __debug__ else logging.INFO
        self.logger.setLevel(level)


def watch(logger_name, level=logging.INFO, out=sys.stdout):
    watcher = Watcher(logger_name)
    watcher.watch(level, out)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
import logging

from py2neo import Service, Resource, Node, Rel, Relationship, Subgraph, Path, Finished
from py2neo.cypher.error.core import CypherError, CypherTransactionError
from py2neo.packages.jsonstream import assembled
from py2neo.packages.tart.tables import TextTable
from py2neo.util import is_integer, is_string, xstr


__all__ = ["CypherResource", "CypherTransaction", "RecordListList", "RecordList", "RecordStream",
           "Record", "RecordProducer"]


log = logging.getLogger("py2neo.cypher")


class CypherResource(Service):
    """ Service wrapper for all Cypher functionality, providing access
    to transactions (if available) as well as single statement execution
    and streaming. If the server supports Cypher transactions, these
    will be used for single statement execution; if not, the vanilla
    Cypher endpoint will be used.

    This class will usually be instantiated via a :class:`py2neo.Graph`
    object and will be made available through the
    :attr:`py2neo.Graph.cypher` attribute. Therefore, for single
    statement execution, simply use the :func:`execute` method::

        from py2neo import Graph
        graph = Graph()
        results = graph.cypher.execute("MATCH (n:Person) RETURN n")

    """

    error_class = CypherError

    __instances = {}

    def __new__(cls, uri, transaction_uri=None):
        key = (uri, transaction_uri)
        try:
            inst = cls.__instances[key]
        except KeyError:
            inst = super(CypherResource, cls).__new__(cls)
            inst.bind(uri)
            inst.transaction_uri = transaction_uri
            cls.__instances[key] = inst
        return inst

    def post(self, statement, parameters=None):
        """ Post a Cypher statement to this resource, optionally with
        parameters.

        :param statement: A Cypher statement to execute.
        :param parameters: A dictionary of parameters.
        :rtype: :class:`httpstream.Response`
        """
        payload = {"query": statement}
        if parameters:
            payload["params"] = {}
            for key, value in parameters.items():
                if isinstance(value, (Node, Rel, Relationship)):
                    value = value._id
                payload["params"][key] = value
        log.info("execute %r %r", payload["query"], payload.get("params", {}))
        return self.resource.post(payload)

    def run(self, statement, parameters=None):
        """ Execute a single Cypher statement, ignoring any return value.

        :param statement: A Cypher statement to execute.
        :param parameters: A dictionary of parameters.
        """
        if self.transaction_uri:
            tx = CypherTransaction(self.transaction_uri)
            tx.append(statement, parameters)
            tx.commit()
        else:
            self.post(statement, parameters).close()

    def execute(self, statement, parameters=None):
        """ Execute a single Cypher statement.

        :param statement: A Cypher statement to execute.
        :param parameters: A dictionary of parameters.
        :rtype: :class:`py2neo.cypher.RecordList`
        """
        if self.transaction_uri:
            tx = CypherTransaction(self.transaction_uri)
            tx.append(statement, parameters)
            results = tx.commit()
            return results[0]
        else:
            response = self.post(statement, parameters)
            try:
                return self.graph.hydrate(response.content)
            finally:
                response.close()

    def execute_one(self, statement, parameters=None):
        """ Execute a single Cypher statement and return the value from
        the first column of the first record returned.

        :param statement: A Cypher statement to execute.
        :param parameters: A dictionary of parameters.
        :return: Single return value or :const:`None`.
        """
        if self.transaction_uri:
            tx = CypherTransaction(self.transaction_uri)
            tx.append(statement, parameters)
            results = tx.commit()
            try:
                return results[0][0][0]
            except IndexError:
                return None
        else:
            response = self.post(statement, parameters)
            results = self.graph.hydrate(response.content)
            try:
                return results[0][0]
            except IndexError:
                return None
            finally:
                response.close()

    def stream(self, statement, parameters=None):
        """ Execute the query and return a result iterator.

        :param statement: A Cypher statement to execute.
        :param parameters: A dictionary of parameters.
        :rtype: :class:`py2neo.cypher.RecordStream`
        """
        return RecordStream(self.graph, self.post(statement, parameters))

    def begin(self):
        """ Begin a new transaction.

        :rtype: :class:`py2neo.cypher.CypherTransaction`
        """
        if self.transaction_uri:
            return CypherTransaction(self.transaction_uri)
        else:
            raise NotImplementedError("Transaction support not available from this "
                                      "Neo4j server version")


class CypherTransaction(object):
    """ A transaction is a transient resource that allows multiple Cypher
    statements to be executed within a single server transaction.
    """

    error_class = CypherTransactionError

    def __init__(self, uri):
        log.info("begin")
        self.statements = []
        self.__begin = Resource(uri)
        self.__begin_commit = Resource(uri + "/commit")
        self.__execute = None
        self.__commit = None
        self.__finished = False
        self.graph = self.__begin.graph

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.commit()

    def __assert_unfinished(self):
        if self.__finished:
            raise Finished(self)

    @property
    def _id(self):
        """ The internal server ID of this transaction, if available.
        """
        if self.__execute is None:
            return None
        else:
            return int(self.__execute.uri.path.segments[-1])

    @property
    def finished(self):
        """ Indicates whether or not this transaction has been completed or is
        still open.

        :return: :py:const:`True` if this transaction has finished,
                 :py:const:`False` otherwise
        """
        return self.__finished

    def append(self, statement, parameters=None):
        """ Add a statement to the current queue of statements to be
        executed.

        :param statement: the statement to append
        :param parameters: a dictionary of execution parameters
        """
        self.__assert_unfinished()
        p = {}
        if parameters:
            for key, value in parameters.items():
                if isinstance(value, (Node, Rel, Relationship)):
                    value = value._id
                p[key] = value
        # OrderedDict is used here to avoid statement/parameters ordering bug
        log.info("append %r %r", statement, p)
        self.statements.append(OrderedDict([
            ("statement", statement),
            ("parameters", p),
            ("resultDataContents", ["REST"]),
        ]))

    def post(self, resource):
        self.__assert_unfinished()
        rs = resource.post({"statements": self.statements})
        location = rs.location
        if location:
            self.__execute = Resource(location)
        j = rs.content
        rs.close()
        self.statements = []
        if "commit" in j:
            self.__commit = Resource(j["commit"])
        if "errors" in j:
            errors = j["errors"]
            if len(errors) >= 1:
                error = errors[0]
                raise self.error_class.hydrate(error)
        out = RecordListList()
        for result in j["results"]:
            columns = result["columns"]
            producer = RecordProducer(columns)
            out.append(RecordList(columns, [producer.produce(self.graph.hydrate(r["rest"]))
                                            for r in result["data"]]))
        return out

    def process(self):
        """ Send all pending statements to the server for execution, leaving
        the transaction open for further statements. Along with
        :meth:`append <.CypherTransaction.append>`, this method can be used to
        batch up a number of individual statements into a single HTTP request::

            from py2neo import Graph

            graph = Graph()
            statement = "MERGE (n:Person {name:{N}}) RETURN n"

            tx = graph.cypher.begin()

            def add_names(*names):
                for name in names:
                    tx.append(statement, {"N": name})
                tx.process()

            add_names("Homer", "Marge", "Bart", "Lisa", "Maggie")
            add_names("Peter", "Lois", "Chris", "Meg", "Stewie")

            tx.commit()

        :return: list of results from pending statements
        """
        log.info("process")
        return self.post(self.__execute or self.__begin)

    def commit(self):
        """ Send all pending statements to the server for execution and commit
        the transaction.

        :return: list of results from pending statements
        """
        log.info("commit")
        try:
            return self.post(self.__commit or self.__begin_commit)
        finally:
            self.__finished = True

    def rollback(self):
        """ Rollback the current transaction.
        """
        self.__assert_unfinished()
        log.info("rollback")
        try:
            if self.__execute:
                self.__execute.delete()
        finally:
            self.__finished = True


class RecordListList(list):
    """ Container for multiple RecordList instances that presents a more
    consistent representation.
    """

    def __repr__(self):
        out = []
        for i in self:
            out.append(repr(i))
        return "\n".join(out)


class RecordList(object):
    """ A list of records returned from the execution of a Cypher statement.
    """

    @classmethod
    def hydrate(cls, data, graph):
        columns = data["columns"]
        rows = data["data"]
        producer = RecordProducer(columns)
        return cls(columns, [producer.produce(graph.hydrate(row)) for row in rows])

    def __init__(self, columns, records):
        log.info("result %r %r", columns, len(records))
        self.columns = columns
        self.records = records

    def __repr__(self):
        out = ""
        if self.columns:
            table = TextTable([None] + self.columns, border=True)
            for i, record in enumerate(self.records):
                table.append([i + 1] + list(record))
            out = repr(table)
        return out

    def __len__(self):
        return len(self.records)

    def __getitem__(self, item):
        return self.records[item]

    def __iter__(self):
        return iter(self.records)

    def to_subgraph(self):
        """ Convert a RecordList into a Subgraph.
        """
        entities = []
        for record in self.records:
            for value in record:
                if isinstance(value, (Node, Path)):
                    entities.append(value)
        return Subgraph(*entities)


class RecordStream(object):
    """ An accessor for a sequence of records yielded by a streamed Cypher statement.

    ::

        for record in graph.cypher.stream("START n=node(*) RETURN n LIMIT 10")
            print record[0]

    Each record returned is cast into a :py:class:`namedtuple` with names
    derived from the resulting column names.

    .. note ::
        Results are available as returned from the server and are decoded
        incrementally. This means that there is no need to wait for the
        entire response to be received before processing can occur.
    """

    def __init__(self, graph, response):
        self.graph = graph
        self.__response = response
        self.__response_item = self.__response_iterator()
        self.columns = next(self.__response_item)
        log.info("stream %r", self.columns)

    def __response_iterator(self):
        producer = None
        columns = []
        record_data = None
        for key, value in self.__response:
            key_len = len(key)
            if key_len > 0:
                section = key[0]
                if section == "columns":
                    if key_len > 1:
                        columns.append(value)
                elif section == "data":
                    if key_len == 1:
                        producer = RecordProducer(columns)
                        yield tuple(columns)
                    elif key_len == 2:
                        if record_data is not None:
                            yield producer.produce(self.graph.hydrate(assembled(record_data)))
                        record_data = []
                    else:
                        record_data.append((key[2:], value))
        if record_data is not None:
            yield producer.produce(self.graph.hydrate(assembled(record_data)))
        self.close()

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.__response_item)

    def next(self):
        return self.__next__()

    def close(self):
        """ Close results and free resources.
        """
        self.__response.close()


class Record(object):
    """ A simple object containing values from a single row of a Cypher
    result. Each value can be retrieved by column position or name,
    supplied as either an index key or an attribute name.

    Consider the record below::

           | person                     | name
        ---+----------------------------+-------
         1 | (n1:Person {name:"Alice"}) | Alice

    If this record is named ``r``, the following expressions
    are equivalent and will return the value ``'Alice'``::

        r[1]
        r["name"]
        r.name

    """

    __producer__ = None

    def __init__(self, values):
        columns = self.__producer__.columns
        for i, column in enumerate(columns):
            setattr(self, column, values[i])

    def __repr__(self):
        out = ""
        columns = self.__producer__.columns
        if columns:
            table = TextTable(columns, border=True)
            table.append([getattr(self, column) for column in columns])
            out = repr(table)
        return out

    def __eq__(self, other):
        return vars(self) == vars(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return self.__producer__.__len__()

    def __getitem__(self, item):
        if is_string(item):
            return getattr(self, item)
        elif is_integer(item):
            return getattr(self, self.__producer__.columns[item])
        else:
            raise LookupError(item)


class RecordProducer(object):

    def __init__(self, columns):
        self.__columns = tuple(column for column in columns if not column.startswith("_"))
        self.__len = len(self.__columns)
        dct = dict.fromkeys(self.__columns)
        dct["__producer__"] = self
        self.__type = type(xstr("Record"), (Record,), dct)

    def __repr__(self):
        return "RecordProducer(columns=%r)" % (self.__columns,)

    def __len__(self):
        return self.__len

    @property
    def columns(self):
        return self.__columns

    def produce(self, values):
        """ Produce a record from a set of values.
        """
        return self.__type(values)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2014, Simon Harrison
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


try:
    from shapely.geos import ReadingError
    from shapely.wkt import loads as wkt_from_string_loader
    from shapely.wkt import dumps
except ImportError:
    print("Please install extension requirements. See README.rst.")

from py2neo import Node, ServerPlugin
from py2neo.error import GraphError
from py2neo.packages.jsonstream import assembled
from py2neo.ext.spatial.exceptions import (
    GeometryExistsError, InvalidWKTError, LayerNotFoundError,
    NodeNotFoundError)
from py2neo.ext.spatial.util import parse_lat_long


EXTENSION_NAME = "SpatialPlugin"
WKT_PROPERTY = "wkt"
NAME_PROPERTY = "_py2neo_geometry_name"

# wkt index config for the contrib spatial extension
EXTENSION_CONFIG = {
    'format': 'WKT',
    'nodePropertyName': WKT_PROPERTY,
}

# shape identifiers
MULTIPOLYGON = 'MultiPolygon'
POINT = 'Point'

# a baseline label so we can retieve all data added via this extension
DEFAULT_LABEL = 'py2neo_spatial'


class Spatial(ServerPlugin):
    """ An API extension to py2neo for WKT type GIS operations.

    .. note::

        An OSMLayer (Open Street Map Layer ) is also possible, but because one
        cannot be created over REST, only via the Java OSMImporter api, there
        is no api to query or manipulate such a layer here.

    """
    def __init__(self, graph):
        super(Spatial, self).__init__(graph, EXTENSION_NAME)

    def _get_data_nodes(self, geometry_nodes):
        ids = [n._id for n in geometry_nodes]
        query = (
            "MATCH (n)-[:LOCATES]->(data_node) "
            "WHERE id(n) IN {ids} RETURN data_node "
        )

        params = {
            'ids': ids,
        }

        results = self.graph.cypher.execute(query, params)
        nodes = [record[0] for record in results]

        return nodes

    def _execute_spatial_request(self, resource, spatial_payload):
        try:
            json_stream = resource.post(spatial_payload)
        except GraphError as exc:
            if 'NullPointerException' in exc.full_name:
                # no results leads to a NullPointerException.
                # this is probably a bug on the Java side, but this
                # happens with some resources and must be managed.
                return []
            raise

        if json_stream.status_code == 204:
            # no content
            return []

        geometry_nodes = map(Node.hydrate, assembled(json_stream))
        nodes = self._get_data_nodes(geometry_nodes)

        return nodes

    def _get_shape_from_wkt(self, wkt_string):
        try:
            shape = wkt_from_string_loader(wkt_string)
        except ReadingError:
            raise InvalidWKTError(
                'Invalid WKT:{}'.format(wkt_string)
            )

        return shape

    def _get_wkt_from_shape(self, shape):
        if shape.type == POINT:
            # shapely float precision errors break cypher matches!
            wkt = dumps(shape, rounding_precision=8)
        else:
            wkt = shape.wkt

        return wkt

    def _geometry_exists(self, shape, geometry_name):
        match = "MATCH (n:{label} ".format(label=shape.type)
        query = match + "{_py2neo_geometry_name:{geometry_name}}) RETURN n"
        params = {
            'geometry_name': geometry_name,
            'wkt': shape.wkt,
        }

        exists = self.graph.cypher.execute(query, params)

        return bool(exists)

    def _layer_exists(self, layer_name):
        query = "MATCH (l {layer:{layer_name}})<-[:LAYER]-() RETURN l"
        params = {'layer_name': layer_name}
        exists = self.graph.cypher.execute(query, params)

        return bool(exists)

    def create_layer(self, layer_name):
        """ Create a Layer to add geometries to. If a Layer with the
        name property value of ``layer_name`` already exists, nothing
        happens.

        """
        resource = self.resources['addEditableLayer']
        spatial_data = dict(layer=layer_name, **EXTENSION_CONFIG)
        raw = resource.post(spatial_data)
        layer = assembled(raw)

        return layer

    def get_layer(self, layer_name):
        resource = self.resources['getLayer']
        spatial_data = dict(layer=layer_name, **EXTENSION_CONFIG)
        raw = resource.post(spatial_data)
        layer = assembled(raw)

        return layer

    def delete_layer(self, layer_name):
        """ Remove a GIS map Layer.

        .. note ::
            There is no Resource for this action so in the meantime we
            use the core py2neo cypher api.

        This will remove a representation of a GIS map Layer from the Neo4j
        data store - it will not remove any nodes you may have added to it.

        The operation removes the layer data from the internal GIS R-tree
        model and removes the layer's label from all nodes that exist on it.
        It does not destroy any Nodes on the DB - use the standard py2neo
        library for these actions.

        :Raises:
            LayerNotFoundError if the index does not exist.

        """
        if not self._layer_exists(layer_name):
            raise LayerNotFoundError(
                'Layer Not Found: "{}"'.format(layer_name)
            )

        graph = self.graph

        # remove labels and properties on Nodes relating to this layer
        query = (
            "MATCH (n:{layer_name}) "
            "REMOVE n:{default_label} "
            "REMOVE n:{layer_name} "
            "REMOVE n:{point_label} "
            "REMOVE n:{multipolygon_label} "
            "REMOVE n.{internal_name}".format(
                layer_name=layer_name, default_label=DEFAULT_LABEL,
                point_label=POINT, multipolygon_label=MULTIPOLYGON,
                internal_name=NAME_PROPERTY
        ))

        graph.cypher.execute(query)

        # remove the bounding box, metadata and root from the rtree index
        query = (
            "MATCH (l { layer:{layer_name} })-[r_layer:LAYER]-(), "
            "(metadata)<-[r_meta:RTREE_METADATA]-(l), "
            "()-[locate_rel:LOCATES]-(geometry_node)-[r_ref:RTREE_REFERENCE]-"
            "(bounding_box)-[r_root:RTREE_ROOT]-(l) "
            "DELETE locate_rel, r_meta, r_layer, r_ref, r_root, "
            "metadata, geometry_node, bounding_box, l"
        )

        params = {
            'layer_name': layer_name
        }

        graph.cypher.execute(query, params)

    def create_geometry(
            self, geometry_name, wkt_string, layer_name, labels=None,
            node_id=None):
        """ Create a geometry of type Well Known Text (WKT).

        Internally this creates a node in your graph with a wkt property
        and also adds it to a GIS map layer (an index). Optionaly add
        Labels to the Node.

        :Params:
            geometry_name : str
                A unique name for the geometry.
            wkt_string : str
                A Well Known Text string of any geometry
            layer_name : str
                The name of the layer to add the geometry to.
            labels : list
                Optional list of Label names to apply to the geometry Node.

            node_id : int
                Optional - the internal ID used by neo4j to uniquely identify
                a Node. When provided, an update operation in the application
                graph will be carried out instead of the usual create, making
                this Node spatially aware by adding a 'wkt' property to it.
                It is then added to the Layer (indexed) as normal.

        :Raises:
            LayerNotFoundError if the index does not exist.
            InvalidWKTError if the WKT cannot be read.

        """
        if not self._layer_exists(layer_name):
            raise LayerNotFoundError(
                'Layer Not Found: "{0}".',
                'Use ``create_layer(layer_name="{0}"")`` first.'.format(
                    layer_name)
            )

        shape = self._get_shape_from_wkt(wkt_string)

        if self._geometry_exists(shape, geometry_name):
            raise GeometryExistsError(
                'geometry already exists. ignoring request.'
            )

        graph = self.graph
        resource = self.resources['addGeometryWKTToLayer']

        labels = labels or []
        labels.extend([DEFAULT_LABEL, layer_name, shape.type])
        wkt = self._get_wkt_from_shape(shape)

        params = {
            NAME_PROPERTY: geometry_name,
        }

        if node_id:
            query = 'MATCH (n) WHERE id(n) = {node_id} RETURN n'
            params = {'node_id': node_id}
            results = graph.cypher.execute(query, params)
            if not results:
                raise NodeNotFoundError(
                    'Node not found: "{}"'.format(node_id)
                )

            record = results[0]
            node = record[0]
            node[NAME_PROPERTY] = geometry_name
            node.add_labels(*tuple(labels))
            node.push()

        else:
            node = Node(*labels, **params)
            graph.create(node)

        spatial_data = {
            'geometry': wkt,
            'layer': layer_name
        }

        resource.post(spatial_data)

        # now relate the geometry to our application node
        query = (
            "MATCH (l { layer:{layer_name} })<-[r_layer:LAYER]-"
            "(root { name:'spatial_root' }), "
            "(bbox)-[r_root:RTREE_ROOT]-(l), "
            "(geometry_node)-[r_ref:RTREE_REFERENCE]-(bbox), "
            "(application_node { _py2neo_geometry_name:{geometry_name} }) "
            "WHERE geometry_node.wkt = {wkt} "
            "CREATE UNIQUE (geometry_node)-[:LOCATES]->(application_node)"
        )

        params = {
            'wkt': wkt,
            'layer_name': layer_name,
            'geometry_name': geometry_name,
        }

        graph.cypher.execute(query, params)

    def delete_geometry(self, geometry_name, wkt_string, layer_name):
        """ Remove a geometry node from a GIS map layer.

        .. note ::
            There is no Resource for this action so in the meantime we
            use the core py2neo cypher api.

        :Params:
            geometry_name : str
                The unique name of the geometry to delete.
            wkt_string : str
                A Well Known Text string of any geometry
            layer_name : str
                The name of the layer/index to remove the geometry from.

        :Raises:
            LayerNotFoundError if the index does not exist.
            InvalidWKTError if the WKT cannot be read.

        """
        if not self._layer_exists(layer_name):
            raise LayerNotFoundError(
                'Layer Not Found: "{}"'.format(layer_name)
            )

        graph = self.graph
        shape = self._get_shape_from_wkt(wkt_string)

        # remove the node from the graph
        match = "MATCH (n:{label}".format(label=shape.type)
        query = match + (
            "{ _py2neo_geometry_name:{geometry_name} }) "
            "OPTIONAL MATCH n<-[r]-() "
            "DELETE r, n"
        )
        params = {
            'label': shape.type,
            'geometry_name': geometry_name,
        }
        graph.cypher.execute(query, params)

        # tidy up the index. This will remove the node,
        # it's bounding box node, and the relationship between them.
        query = (
            "MATCH (l { layer:{layer_name} }), "
            "(n { wkt:{wkt} })-[ref:RTREE_REFERENCE]-() "
            "DELETE ref, n"
        )
        params = {
            'layer_name': layer_name,
            'wkt': shape.wkt,
        }
        graph.cypher.execute(query, params)

    def update_geometry(self, geometry_name, wkt_string):
        """ Update the geometry on a Node.

        :Params:
            wkt_string : str
                A Well Known Text string of any geometry

        :Raises:
            InvalidWKTError if the WKT cannot be read.

        """
        graph = self.graph
        resource = self.resources['updateGeometryFromWKT']
        shape = self._get_shape_from_wkt(wkt_string)

        query = (
            "MATCH (application_node {_py2neo_geometry_name:{geometry_name}}),"
            "(application_node)<-[:LOCATES]-(geometry_node)"
            "<-[:RTREE_REFERENCE]-()<-[:RTREE_ROOT]-(layer_node) "
            "RETURN geometry_node, layer_node"
        )
        params = {
            'geometry_name': geometry_name,
        }

        result = graph.cypher.execute(query, params)
        if not result:
            raise NodeNotFoundError(
                'Cannot update Node - Node not found: "{}"'.format(
                    geometry_name)
            )

        records = result[0]
        geometry_node, layer_node = records
        geometry_node_id = int(geometry_node.uri.path.segments[-1])
        geometry_layer = layer_node.get_properties()['layer']

        spatial_data = {
            'geometry': shape.wkt,
            'geometryNodeId': geometry_node_id,
            'layer': geometry_layer,
        }

        # update the geometry node
        self._execute_spatial_request(resource, spatial_data)

    def find_within_distance(self, layer_name, coords, distance):
        """ Find all points of interest (poi) within a given distance from
        a lat-lon location coord.

        :Params:
            layer_name : str
                The name of the layer/index to remove the geometry from.
            coords : tuple
                WGS84 (EPSG 4326) lat, lon pair
                Latitude is a decimal number between -90.0 and 90.0
                Longitude is a decimal number between -180.0 and 180.0
            distance : int
                The radius of the search area in Kilometres (km)

        :Raises:
            LayerNotFoundError if the index does not exist.

        :Returns:
            a list of all matched nodes

        """
        if not self._layer_exists(layer_name):
            raise LayerNotFoundError(
                'Layer Not Found: "{}"'.format(layer_name)
            )

        resource = self.resources['findGeometriesWithinDistance']
        shape = parse_lat_long(coords)
        spatial_data = {
            'pointX': shape.x,
            'pointY': shape.y,
            'layer': layer_name,
            'distanceInKm': distance,
        }

        nodes = self._execute_spatial_request(resource, spatial_data)
        return nodes

    def find_closest_geometries(self, coords):
        """ Find the "closest" points of interest (poi) accross *all* layers
        from a given lat-lon location coord.

        :Params:
            coords : tuple
                WGS84 (EPSG 4326) lat, lon pair
                Latitude is a decimal number between -90.0 and 90.0
                Longitude is a decimal number between -180.0 and 180.0

        :Returns:
            a list of all matched nodes

        """
        resource = self.resources['findClosestGeometries']
        shape = parse_lat_long(coords)
        query = "MATCH (r { name:'spatial_root' }), (r)-[:LAYER]->(n) RETURN n"
        results = self.graph.cypher.execute(query)

        spatial_data = {
            'pointX': shape.x,
            'pointY': shape.y,
            # this appears to be handled more like a 'tolerance', as increasing
            # the value even slightly returns data from hundreds of kms away.
            # TODO: raise failing spatial PR with Spatial API.
            'distanceInKm': 4,
        }

        pois = []
        for record in results:
            node = record[0]
            node_properties = node.get_properties()
            layer_name = node_properties['layer']
            spatial_data['layer'] = layer_name
            nodes = self._execute_spatial_request(resource, spatial_data)
            pois.extend(nodes)

        return pois

    def find_within_bounding_box(self, layer_name, minx, miny, maxx, maxy):
        """ Find the points of interest from a given layer enclosed by a
        bounding box.

        The bounding box is definded by the lat-longs of the bottom left and
        the top right, essentially::

            bbox = (min Longitude, min Latitude, max Longitude, max Latitude)

        :Params:
            layer_name : str
                The name of the layer/index to remove the geometry from.
            minx : Decimal
                longitude of the bottom-left corner
            miny : Decimal
                latitude of the bottom-left corner
            maxx : Decimal
                longitude of the top-right corner
            minx : Decimal
                latitude of the top-right corner

        """
        resource = self.resources['findGeometriesInBBox']
        spatial_data = {
            'layer': layer_name,
            'minx': minx,
            'maxx': maxx,
            'miny': miny,
            'maxy': maxy,
        }

        nodes = self._execute_spatial_request(resource, spatial_data)
        return nodes



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2013-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys


if sys.version_info >= (3,):

    is_integer = lambda x: isinstance(x, int)
    is_numeric = lambda x: isinstance(x, (int, float, complex))
    is_string = lambda x: isinstance(x, str)

    def bstr(s, encoding="utf-8"):
        if isinstance(s, bytes):
            return s
        elif isinstance(s, bytearray):
            return bytes(s)
        elif isinstance(s, str):
            return bytes(s, encoding)
        else:
            return bytes(str(s), encoding)

    def ustr(s, encoding="utf-8"):
        """ Convert argument to unicode string.
        """
        if isinstance(s, str):
            return s
        try:
            return s.decode(encoding)
        except AttributeError:
            return str(s)

    def xstr(s, encoding="utf-8"):
        """ Convert argument to string type returned by __str__.
        """
        return ustr(s, encoding)

else:

    is_integer = lambda x: isinstance(x, (int, long))
    is_numeric = lambda x: isinstance(x, (int, float, long, complex))
    is_string = lambda x: isinstance(x, (str, unicode))

    def bstr(s, encoding="utf-8"):
        if isinstance(s, bytes):
            return s
        elif isinstance(s, bytearray):
            return bytes(s)
        elif isinstance(s, unicode):
            return s.encode(encoding)
        else:
            return str(s)

    def ustr(s, encoding="utf-8"):
        """ Convert argument to unicode string.
        """
        if isinstance(s, str):
            return s.decode(encoding)
        else:
            return unicode(s)

    def xstr(s, encoding="utf-8"):
        """ Convert argument to string type returned by __str__.
        """
        if isinstance(s, str):
            return s
        else:
            return unicode(s).encode(encoding)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import division, unicode_literals

import base64
from io import StringIO
import re
from warnings import warn
import webbrowser

from py2neo import __version__
from py2neo.error import BindError, GraphError, JoinError
from py2neo.packages.httpstream import http, ClientError, ServerError, \
    Resource as _Resource, ResourceTemplate as _ResourceTemplate
from py2neo.packages.httpstream.http import JSONResponse, user_agent
from py2neo.packages.httpstream.numbers import NOT_FOUND
from py2neo.packages.httpstream.packages.urimagic import URI
from py2neo.types import cast_property
from py2neo.util import is_collection, is_integer, round_robin, ustr, version_tuple, \
    raise_from, xstr, ThreadLocalWeakValueDictionary


__all__ = ["Graph", "Node", "Relationship", "Path", "NodePointer", "Rel", "Rev", "Subgraph",
           "ServiceRoot", "PropertySet", "LabelSet", "PropertyContainer",
           "authenticate", "rewrite", "ServerPlugin", "UnmanagedExtension",
           "Service", "Resource", "ResourceTemplate"]


DEFAULT_SCHEME = "http"
DEFAULT_HOST = "localhost"
DEFAULT_PORT = 7474
DEFAULT_HOST_PORT = "{0}:{1}".format(DEFAULT_HOST, DEFAULT_PORT)

PRODUCT = ("py2neo", __version__)

NON_ALPHA_NUM = re.compile("[^0-9A-Za-z_]")
SIMPLE_NAME = re.compile(r"[A-Za-z_][0-9A-Za-z_]*")

http.default_encoding = "UTF-8"

_headers = {
    None: [
        ("User-Agent", user_agent(PRODUCT)),
        ("X-Stream", "true"),
    ],
}

_http_rewrites = {}


def _add_header(key, value, host_port=None):
    """ Add an HTTP header to be sent with all requests if no `host_port`
    is provided or only to those matching the value supplied otherwise.
    """
    if host_port in _headers:
        _headers[host_port].append((key, value))
    else:
        _headers[host_port] = [(key, value)]


def _get_headers(host_port):
    """Fetch all HTTP headers relevant to the `host_port` provided.
    """
    uri_headers = {}
    for n, headers in _headers.items():
        if n is None or n == host_port:
            uri_headers.update(headers)
    return uri_headers


def authenticate(host_port, user_name, password):
    """ Set HTTP basic authentication values for specified `host_port`. The
    code below shows a simple example::

        from py2neo import authenticate, Graph

        # set up authentication parameters
        authenticate("camelot:7474", "arthur", "excalibur")

        # connect to authenticated graph database
        graph = Graph("http://camelot:7474/db/data/")

    Note: a `host_port` can be either a server name or a server name and port
    number but must match exactly that used within the Graph
    URI.

    :arg host_port: the host and optional port requiring authentication
        (e.g. "bigserver", "camelot:7474")
    :arg user_name: the user name to authenticate as
    :arg password: the password
    """
    credentials = (user_name + ":" + password).encode("UTF-8")
    value = "Basic " + base64.b64encode(credentials).decode("ASCII")
    _add_header("Authorization", value, host_port=host_port)


def rewrite(from_scheme_host_port, to_scheme_host_port):
    """ Automatically rewrite all URIs directed to the scheme, host and port
    specified in `from_scheme_host_port` to that specified in
    `to_scheme_host_port`.

    As an example::

        from py2neo import rewrite
        # implicitly convert all URIs beginning with <http://localhost:7474>
        # to instead use <https://dbserver:9999>
        rewrite(("http", "localhost", 7474), ("https", "dbserver", 9999))

    If `to_scheme_host_port` is :py:const:`None` then any rewrite rule for
    `from_scheme_host_port` is removed.

    This facility is primarily intended for use by database servers behind
    proxies which are unaware of their externally visible network address.
    """
    global _http_rewrites
    if to_scheme_host_port is None:
        try:
            del _http_rewrites[from_scheme_host_port]
        except KeyError:
            pass
    else:
        _http_rewrites[from_scheme_host_port] = to_scheme_host_port


class Resource(_Resource):
    """ Base class for all local resources mapped to remote counterparts.
    """

    #: The class of error raised by failure responses from this resource.
    error_class = GraphError

    def __init__(self, uri, metadata=None):
        uri = URI(uri)
        scheme_host_port = (uri.scheme, uri.host, uri.port)
        if scheme_host_port in _http_rewrites:
            scheme_host_port = _http_rewrites[scheme_host_port]
            # This is fine - it's all my code anyway...
            uri._URI__set_scheme(scheme_host_port[0])
            uri._URI__set_authority("{0}:{1}".format(scheme_host_port[1],
                                                     scheme_host_port[2]))
        if uri.user_info:
            authenticate(uri.host_port, *uri.user_info.partition(":")[0::2])
        self._resource = _Resource.__init__(self, uri)
        #self._subresources = {}
        self.__headers = _get_headers(self.__uri__.host_port)
        self.__base = super(Resource, self)
        if metadata is None:
            self.__initial_metadata = None
        else:
            self.__initial_metadata = dict(metadata)
        self.__last_get_response = None

        uri = uri.string
        service_root_uri = uri[:uri.find("/", uri.find("//") + 2)] + "/"
        if service_root_uri == uri:
            self.__service_root = self
        else:
            self.__service_root = ServiceRoot(service_root_uri)
        self.__ref = NotImplemented

    @property
    def graph(self):
        """ The parent graph of this resource.

        :rtype: :class:`.Graph`
        """
        return self.__service_root.graph

    @property
    def headers(self):
        """ The HTTP headers sent with this resource.
        """
        return self.__headers

    @property
    def metadata(self):
        """ Metadata received in the last HTTP response.
        """
        if self.__last_get_response is None:
            if self.__initial_metadata is not None:
                return self.__initial_metadata
            self.get()
        return self.__last_get_response.content

    @property
    def ref(self):
        """ The URI of this resource relative to its graph.

        :rtype: string
        """
        if self.__ref is NotImplemented:
            self_uri = self.uri.string
            graph_uri = self.graph.uri.string
            self.__ref = self_uri[len(graph_uri):]
        return self.__ref

    @property
    def service_root(self):
        """ The root service associated with this resource.

        :return: :class:`.ServiceRoot`
        """
        return self.__service_root

    def get(self, headers=None, redirect_limit=5, **kwargs):
        """ Perform an HTTP GET to this resource.

        :arg headers: Extra headers to pass in the request.
        :arg redirect_limit: Maximum number of times to follow redirects.
        :arg kwargs: Other arguments to pass to the underlying `httpstream` method.
        :rtype: :class:`httpstream.Response`
        :raises: :class:`py2neo.GraphError`
        """
        headers = dict(headers or {})
        headers.update(self.__headers)
        kwargs.update(cache=True)
        try:
            response = self.__base.get(headers=headers, redirect_limit=redirect_limit, **kwargs)
        except (ClientError, ServerError) as error:
            if isinstance(error, JSONResponse):
                content = dict(error.content, request=error.request, response=error)
            else:
                content = {}
            message = content.pop("message", "HTTP GET returned response %s" % error.status_code)
            raise_from(self.error_class(message, **content), error)
        else:
            self.__last_get_response = response
            return response

    def put(self, body=None, headers=None, **kwargs):
        """ Perform an HTTP PUT to this resource.

        :arg body: The payload of this request.
        :arg headers: Extra headers to pass in the request.
        :arg kwargs: Other arguments to pass to the underlying `httpstream` method.
        :rtype: :class:`httpstream.Response`
        :raises: :class:`py2neo.GraphError`
        """
        headers = dict(headers or {})
        headers.update(self.__headers)
        try:
            response = self.__base.put(body, headers, **kwargs)
        except (ClientError, ServerError) as error:
            if isinstance(error, JSONResponse):
                content = dict(error.content, request=error.request, response=error)
            else:
                content = {}
            message = content.pop("message", "HTTP PUT returned response %s" % error.status_code)
            raise_from(self.error_class(message, **content), error)
        else:
            return response

    def post(self, body=None, headers=None, **kwargs):
        """ Perform an HTTP POST to this resource.

        :arg body: The payload of this request.
        :arg headers: Extra headers to pass in the request.
        :arg kwargs: Other arguments to pass to the underlying `httpstream` method.
        :rtype: :class:`httpstream.Response`
        :raises: :class:`py2neo.GraphError`
        """
        headers = dict(headers or {})
        headers.update(self.__headers)
        try:
            response = self.__base.post(body, headers, **kwargs)
        except (ClientError, ServerError) as error:
            if isinstance(error, JSONResponse):
                content = dict(error.content, request=error.request, response=error)
            else:
                content = {}
            message = content.pop("message", "HTTP POST returned response %s" % error.status_code)
            raise_from(self.error_class(message, **content), error)
        else:
            return response

    def delete(self, headers=None, **kwargs):
        """ Perform an HTTP DELETE to this resource.

        :arg headers: Extra headers to pass in the request.
        :arg kwargs: Other arguments to pass to the underlying `httpstream` method.
        :rtype: :class:`httpstream.Response`
        :raises: :class:`py2neo.GraphError`
        """
        headers = dict(headers or {})
        headers.update(self.__headers)
        try:
            response = self.__base.delete(headers, **kwargs)
        except (ClientError, ServerError) as error:
            if isinstance(error, JSONResponse):
                content = dict(error.content, request=error.request, response=error)
            else:
                content = {}
            message = content.pop("message", "HTTP DELETE returned response %s" % error.status_code)
            raise_from(self.error_class(message, **content), error)
        else:
            return response


class ResourceTemplate(_ResourceTemplate):
    """ A factory class for producing :class:`.Resource` objects dynamically
    based on a template URI.
    """

    #: The class of error raised by failure responses from resources produced by this template.
    error_class = GraphError

    def expand(self, **values):
        """ Produce a resource instance by substituting values into the
        stored template URI.

        :arg values: A set of named values to plug into the template URI.
        :rtype: :class:`.Resource`
        """
        resource = Resource(self.uri_template.expand(**values))
        resource.error_class = self.error_class
        return resource


class Service(object):
    """ Base class for objects that can be optionally bound to a remote resource. This
    class is essentially a container for a :class:`.Resource` instance.
    """

    #: The class of error raised by failure responses from the contained resource.
    error_class = GraphError

    __resource__ = None

    def bind(self, uri, metadata=None):
        """ Associate this «class.lower» with a remote resource.

        :arg uri: The URI identifying the remote resource to which to bind.
        :arg metadata: Dictionary of initial metadata to attach to the contained resource.

        """
        if "{" in uri and "}" in uri:
            if metadata:
                raise ValueError("Initial metadata cannot be passed to a resource template")
            self.__resource__ = ResourceTemplate(uri)
        else:
            self.__resource__ = Resource(uri, metadata)
        self.__resource__.error_class = self.error_class

    @property
    def bound(self):
        """ :const:`True` if this object is bound to a remote resource,
        :const:`False` otherwise.
        """
        return self.__resource__ is not None

    @property
    def graph(self):
        """ The graph associated with the remote resource.

        :rtype: :class:`.Graph`
        """
        return self.service_root.graph

    @property
    def ref(self):
        """ The URI of the remote resource relative to its graph.

        :rtype: string
        """
        return self.resource.ref

    @property
    def resource(self):
        """ The remote resource to which this object is bound.

        :rtype: :class:`.Resource`
        :raises: :class:`py2neo.BindError`
        """
        if self.bound:
            return self.__resource__
        else:
            raise BindError("Local entity is not bound to a remote entity")

    @property
    def service_root(self):
        """ The root service associated with the remote resource.

        :return: :class:`.ServiceRoot`
        """
        return self.resource.service_root

    def unbind(self):
        """ Detach this object from any remote resource.
        """
        self.__resource__ = None

    @property
    def uri(self):
        """ The full URI of the remote resource.
        """
        resource = self.resource
        try:
            return resource.uri
        except AttributeError:
            return resource.uri_template


class ServiceRoot(object):
    """ Wrapper for the base REST resource exposed by a running Neo4j
    server, corresponding to the ``/`` URI.
    """

    #: The URI for a Neo4j service with default configuration.
    DEFAULT_URI = "{0}://{1}/".format(DEFAULT_SCHEME, DEFAULT_HOST_PORT)

    __instances = {}

    __graph = None

    def __new__(cls, uri=None):
        if uri is None:
            uri = cls.DEFAULT_URI
        if not uri.endswith("/"):
            uri += "/"
        try:
            inst = cls.__instances[uri]
        except KeyError:
            inst = super(ServiceRoot, cls).__new__(cls)
            inst.__resource = Resource(uri)
            inst.__graph = None
            cls.__instances[uri] = inst
        return inst

    @property
    def graph(self):
        """ The graph exposed by this service.

        :rtype: :class:`.Graph`
        """
        if self.__graph is None:
            self.__graph = Graph(self.resource.metadata["data"])
        return self.__graph

    @property
    def resource(self):
        """ The contained resource object for this instance.

        :rtype: :class:`py2neo.Resource`
        """
        return self.__resource

    @property
    def uri(self):
        """ The full URI of the contained resource.
        """
        return self.resource.uri


class Graph(Service):
    """ The `Graph` class provides a wrapper around the
    `REST API <http://docs.neo4j.org/chunked/stable/rest-api.html>`_ exposed
    by a running Neo4j database server and is identified by the base URI
    of the graph database. If no URI is specified, a default value of
    `http://localhost:7474/db/data/` is assumed; therefore, to connect to a
    local server with default configuration, simply use::

        >>> from py2neo import Graph
        >>> graph = Graph()

    An explicitly specified graph database URI can be passed to the constructor
    as a string::

        >>> other_graph = Graph("http://camelot:1138/db/data/")

    If the database server is behind a proxy that requires HTTP authorisation,
    the relevant criteria can also be specified within the URI::

        >>> secure_graph = Graph("http://arthur:excalibur@camelot:1138/db/data/")

    Once obtained, the `Graph` instance provides direct or indirect access
    to most of the functionality available within py2neo.

    """

    __instances = {}

    __batch = None
    __cypher = None
    __legacy = None
    __schema = None
    __node_labels = None
    __relationship_types = None

    # Auto-sync will be removed in 2.1
    auto_sync_properties = False

    @staticmethod
    def cast(obj):
        """ Cast an general Python object to a graph-specific entity,
        such as a :class:`.Node` or a :class:`.Relationship`.
        """
        if obj is None:
            return None
        elif isinstance(obj, (Node, NodePointer, Path, Rel, Relationship, Rev, Subgraph)):
            return obj
        elif isinstance(obj, dict):
            return Node.cast(obj)
        elif isinstance(obj, tuple):
            return Relationship.cast(obj)
        else:
            raise TypeError(obj)

    def __new__(cls, uri=None):
        if uri is None:
            uri = ServiceRoot().graph.uri.string
        if not uri.endswith("/"):
            uri += "/"
        key = (cls, uri)
        try:
            inst = cls.__instances[key]
        except KeyError:
            inst = super(Graph, cls).__new__(cls)
            inst.bind(uri)
            cls.__instances[key] = inst
        return inst

    def __repr__(self):
        return "<Graph uri=%r>" % self.uri.string

    def __hash__(self):
        return hash(self.uri)

    def __len__(self):
        return self.size

    def __bool__(self):
        return True

    def __nonzero__(self):
        return True

    def __contains__(self, entity):
        return entity.bound and entity.uri.string.startswith(entity.uri.string)

    @property
    def batch(self):
        """ A :class:`py2neo.batch.BatchResource` instance attached to this
        graph. This resource exposes methods for submitting iterable
        collections of :class:`py2neo.batch.Job` objects to the server and
        will often be used indirectly via classes such as
        :class:`py2neo.batch.PullBatch` or :class:`py2neo.batch.PushBatch`.

        :rtype: :class:`py2neo.cypher.BatchResource`

        """
        if self.__batch is None:
            from py2neo.batch import BatchResource
            self.__batch = BatchResource(self.resource.metadata["batch"])
        return self.__batch

    @property
    def cypher(self):
        """ The Cypher execution resource for this graph providing access to
        all Cypher functionality for the underlying database, both simple
        and transactional.

        ::

            >>> from py2neo import Graph
            >>> graph = Graph()
            >>> graph.cypher.execute("CREATE (a:Person {name:{N}})", {"N": "Alice"})

        :rtype: :class:`py2neo.cypher.CypherResource`

        """
        if self.__cypher is None:
            from py2neo.cypher import CypherResource
            metadata = self.resource.metadata
            self.__cypher = CypherResource(metadata["cypher"], metadata.get("transaction"))
        return self.__cypher

    def create(self, *entities):
        """ Create one or more remote nodes, relationships or paths in a
        single transaction. The entity values provided must be either
        existing entity objects (such as nodes or relationships) or values
        that can be cast to them.

        For example, to create a remote node from a local :class:`Node` object::

            from py2neo import Graph, Node
            graph = Graph()
            alice = Node("Person", name="Alice")
            graph.create(alice)

        Then, create a second node and a relationship connecting both nodes::

            german, speaks = graph.create({"name": "German"}, (alice, "SPEAKS", 0))

        This second example shows how :class:`dict` and :class:`tuple` objects
        can also be used to create nodes and relationships respectively. The
        zero value in the relationship tuple references the zeroth item created
        within that transaction, i.e. the "German" node.

        .. note::
            If an object is passed to this method that is already bound to
            a remote entity, that argument will be ignored and nothing will
            be created.

        :arg entities: One or more existing graph entities or values that
                       can be cast to entities.
        :return: A tuple of all entities created (or ignored) of the same
                 length and order as the arguments passed in.
        :rtype: tuple

        .. warning::
            This method will *always* return a tuple, even when creating
            only a single entity. To automatically unpack to a single
            item, append a trailing comma to the variable name on the
            left of the assignment operation.

        """
        from py2neo.cypher.create import CreateStatement
        statement = CreateStatement(self)
        for entity in entities:
            statement.create(entity)
        return statement.execute()

    def create_unique(self, *entities):
        """ Create one or more unique paths or relationships in a single
        transaction. This is similar to :meth:`create` but uses a Cypher
        `CREATE UNIQUE <http://docs.neo4j.org/chunked/stable/query-create-unique.html>`_
        clause to ensure that only relationships that do not already exist are created.
        """
        from py2neo.cypher.create import CreateStatement
        statement = CreateStatement(self)
        for entity in entities:
            statement.create_unique(entity)
        return statement.execute()

    def delete(self, *entities):
        """ Delete one or more nodes, relationships and/or paths.
        """
        from py2neo.cypher.delete import DeleteStatement
        statement = DeleteStatement(self)
        for entity in entities:
            statement.delete(entity)
        return statement.execute()

    def delete_all(self):
        """ Delete all nodes and relationships from the graph.

        .. warning::
            This method will permanently remove **all** nodes and relationships
            from the graph and cannot be undone.
        """
        from py2neo.batch import WriteBatch, CypherJob
        batch = WriteBatch(self)
        batch.append(CypherJob("START r=rel(*) DELETE r"))
        batch.append(CypherJob("START n=node(*) DELETE n"))
        batch.run()

    def find(self, label, property_key=None, property_value=None, limit=None):
        """ Iterate through a set of labelled nodes, optionally filtering
        by property key and value
        """
        if not label:
            raise ValueError("Empty label")
        from py2neo.cypher.lang import cypher_escape
        if property_key is None:
            statement = "MATCH (n:%s) RETURN n,labels(n)" % cypher_escape(label)
            parameters = {}
        else:
            statement = "MATCH (n:%s {%s:{V}}) RETURN n,labels(n)" % (
                cypher_escape(label), cypher_escape(property_key))
            parameters = {"V": property_value}
        if limit:
            statement += " LIMIT %s" % limit
        response = self.cypher.post(statement, parameters)
        for record in response.content["data"]:
            dehydrated = record[0]
            dehydrated.setdefault("metadata", {})["labels"] = record[1]
            yield self.hydrate(dehydrated)
        response.close()

    def find_one(self, label, property_key=None, property_value=None):
        """ Find a single node by label and optional property. This method is
        intended to be used with a unique constraint and does not fail if more
        than one matching node is found.
        """
        for node in self.find(label, property_key, property_value, limit=1):
            return node

    def hydrate(self, data):
        """ Hydrate a dictionary of data to produce a :class:`.Node`,
        :class:`.Relationship` or other graph object instance. The
        data structure and values expected are those produced by the
        `REST API <http://neo4j.com/docs/stable/rest-api.html>`__.

        :arg data: dictionary of data to hydrate
        
        """
        if isinstance(data, dict):
            if "self" in data:
                if "type" in data:
                    return Relationship.hydrate(data)
                else:
                    return Node.hydrate(data)
            elif "nodes" in data and "relationships" in data:
                if "directions" not in data:
                    from py2neo.batch import Job, Target
                    node_uris = data["nodes"]
                    relationship_uris = data["relationships"]
                    jobs = [Job("GET", Target(uri)) for uri in relationship_uris]
                    directions = []
                    for i, result in enumerate(self.batch.submit(jobs)):
                        rel_data = result.content
                        start = rel_data["start"]
                        end = rel_data["end"]
                        if start == node_uris[i] and end == node_uris[i + 1]:
                            directions.append("->")
                        else:
                            directions.append("<-")
                    data["directions"] = directions
                return Path.hydrate(data)
            elif "columns" in data and "data" in data:
                from py2neo.cypher import RecordList
                return RecordList.hydrate(data, self)
            elif "neo4j_version" in data:
                return self
            elif "exception" in data and "stacktrace" in data:
                message = data.pop("message", "The server returned an error")
                raise GraphError(message, **data)
            else:
                warn("Map literals returned over the Neo4j REST interface are ambiguous "
                     "and may be hydrated as graph objects")
                return data
        elif is_collection(data):
            return type(data)(map(self.hydrate, data))
        else:
            return data

    @property
    def legacy(self):
        """ Sub-resource providing access to legacy functionality.

        :rtype: :class:`py2neo.legacy.LegacyResource`
        """
        if self.__legacy is None:
            from py2neo.legacy import LegacyResource
            self.__legacy = LegacyResource(self.uri.string)
        return self.__legacy

    def match(self, start_node=None, rel_type=None, end_node=None, bidirectional=False, limit=None):
        """ Return an iterator for all relationships matching the
        specified criteria.

        For example, to find all of Alice's friends::

            for rel in graph.match(start_node=alice, rel_type="FRIEND"):
                print(rel.end_node.properties["name"])

        :arg start_node: :attr:`~py2neo.Node.bound` start :class:`~py2neo.Node` to match or
                           :const:`None` if any
        :arg rel_type: type of relationships to match or :const:`None` if any
        :arg end_node: :attr:`~py2neo.Node.bound` end :class:`~py2neo.Node` to match or
                         :const:`None` if any
        :arg bidirectional: :const:`True` if reversed relationships should also be included
        :arg limit: maximum number of relationships to match or :const:`None` if no limit
        :return: matching relationships
        :rtype: generator
        """
        if start_node is None and end_node is None:
            query = "START a=node(*)"
            params = {}
        elif end_node is None:
            query = "START a=node({A})"
            start_node = Node.cast(start_node)
            if not start_node.bound:
                raise TypeError("Nodes for relationship match end points must be bound")
            params = {"A": start_node}
        elif start_node is None:
            query = "START b=node({B})"
            end_node = Node.cast(end_node)
            if not end_node.bound:
                raise TypeError("Nodes for relationship match end points must be bound")
            params = {"B": end_node}
        else:
            query = "START a=node({A}),b=node({B})"
            start_node = Node.cast(start_node)
            end_node = Node.cast(end_node)
            if not start_node.bound or not end_node.bound:
                raise TypeError("Nodes for relationship match end points must be bound")
            params = {"A": start_node, "B": end_node}
        if rel_type is None:
            rel_clause = ""
        elif is_collection(rel_type):
            separator = "|:" if self.neo4j_version >= (2, 0, 0) else "|"
            rel_clause = ":" + separator.join("`{0}`".format(_)
                                              for _ in rel_type)
        else:
            rel_clause = ":`{0}`".format(rel_type)
        if bidirectional:
            query += " MATCH (a)-[r" + rel_clause + "]-(b) RETURN r"
        else:
            query += " MATCH (a)-[r" + rel_clause + "]->(b) RETURN r"
        if limit is not None:
            query += " LIMIT {0}".format(int(limit))
        results = self.cypher.stream(query, params)
        try:
            for result in results:
                yield result.r
        finally:
            results.close()

    def match_one(self, start_node=None, rel_type=None, end_node=None, bidirectional=False):
        """ Return a single relationship matching the
        specified criteria. See :meth:`~py2neo.Graph.match` for
        argument details.
        """
        rels = list(self.match(start_node, rel_type, end_node,
                               bidirectional, 1))
        if rels:
            return rels[0]
        else:
            return None

    def merge(self, label, property_key=None, property_value=None, limit=None):
        """ Match or create a node by label and optional property and return
        all matching nodes.
        """
        if not label:
            raise ValueError("Empty label")
        from py2neo.cypher.lang import cypher_escape
        if property_key is None:
            statement = "MERGE (n:%s) RETURN n,labels(n)" % cypher_escape(label)
            parameters = {}
        else:
            statement = "MERGE (n:%s {%s:{V}}) RETURN n,labels(n)" % (
                cypher_escape(label), cypher_escape(property_key))
            parameters = {"V": property_value}
        if limit:
            statement += " LIMIT %s" % limit
        response = self.cypher.post(statement, parameters)
        for record in response.content["data"]:
            dehydrated = record[0]
            dehydrated.setdefault("metadata", {})["labels"] = record[1]
            yield self.hydrate(dehydrated)
        response.close()

    def merge_one(self, label, property_key=None, property_value=None):
        """ Match or create a node by label and optional property and return a
        single matching node. This method is intended to be used with a unique
        constraint and does not fail if more than one matching node is found.
        """
        for node in self.merge(label, property_key, property_value, limit=1):
            return node

    @property
    def neo4j_version(self):
        """ The database software version as a 4-tuple of (``int``, ``int``,
        ``int``, ``str``).
        """
        return version_tuple(self.resource.metadata["neo4j_version"])

    def node(self, id_):
        """ Fetch a node by ID. This method creates an object representing the
        remote node with the ID specified but fetches no data from the server.
        For this reason, there is no guarantee that the entity returned
        actually exists.
        """
        resource = self.resource.resolve("node/%s" % id_)
        uri_string = resource.uri.string
        try:
            return Node.cache[uri_string]
        except KeyError:
            data = {"self": uri_string}
            return Node.cache.setdefault(uri_string, Node.hydrate(data))

    @property
    def node_labels(self):
        """ The set of node labels currently defined within the graph.
        """
        if not self.supports_node_labels:
            raise NotImplementedError("Node labels not available for this Neo4j server version")
        if self.__node_labels is None:
            self.__node_labels = Resource(self.uri.string + "labels")
        return frozenset(self.__node_labels.get().content)

    def open_browser(self):
        """ Open a page in the default system web browser pointing at
        the Neo4j browser application for this graph.
        """
        webbrowser.open(self.service_root.resource.uri.string)

    @property
    def order(self):
        """ The number of nodes in this graph.
        """
        return self.cypher.execute_one("START n=node(*) RETURN count(n)")

    def pull(self, *entities):
        """ Pull data to one or more entities from their remote counterparts.
        """
        if entities:
            from py2neo.batch.pull import PullBatch
            batch = PullBatch(self)
            for entity in entities:
                batch.append(entity)
            batch.pull()

    def push(self, *entities):
        """ Push data from one or more entities to their remote counterparts.
        """
        if entities:
            from py2neo.batch.push import PushBatch
            batch = PushBatch(self)
            for entity in entities:
                batch.append(entity)
            batch.push()

    def relationship(self, id_):
        """ Fetch a relationship by ID.
        """
        resource = self.resource.resolve("relationship/" + str(id_))
        uri_string = resource.uri.string
        try:
            return Relationship.cache[uri_string]
        except KeyError:
            try:
                return Relationship.cache.setdefault(
                    uri_string, Relationship.hydrate(resource.get().content))
            except ClientError:
                raise ValueError("Relationship with ID %s not found" % id_)

    @property
    def relationship_types(self):
        """ The set of relationship types currently defined within the graph.
        """
        if self.__relationship_types is None:
            self.__relationship_types = Resource(self.uri.string + "relationship/types")
        return frozenset(self.__relationship_types.get().content)

    @property
    def schema(self):
        """ The schema resource for this graph.

        :rtype: :class:`SchemaResource <py2neo.schema.SchemaResource>`
        """
        if self.__schema is None:
            from py2neo.schema import SchemaResource
            self.__schema = SchemaResource(self.uri.string + "schema")
        return self.__schema

    @property
    def size(self):
        """ The number of relationships in this graph.
        """
        return self.cypher.execute_one("START r=rel(*) RETURN count(r)")

    @property
    def supports_cypher_transactions(self):
        """ Indicates whether the server supports explicit Cypher transactions.
        """
        return "transaction" in self.resource.metadata

    @property
    def supports_foreach_pipe(self):
        """ Indicates whether the server supports pipe syntax for FOREACH.
        """
        return self.neo4j_version >= (2, 0)

    @property
    def supports_node_labels(self):
        """ Indicates whether the server supports node labels.
        """
        return self.neo4j_version >= (2, 0)

    @property
    def supports_optional_match(self):
        """ Indicates whether the server supports Cypher OPTIONAL MATCH
        clauses.
        """
        return self.neo4j_version >= (2, 0)

    @property
    def supports_schema_indexes(self):
        """ Indicates whether the server supports schema indexes.
        """
        return self.neo4j_version >= (2, 0)


class PropertySet(Service, dict):
    """ A dict subclass that equates None with a non-existent key and can be
    bound to a remote *properties* resource.
    """

    def __init__(self, iterable=None, **kwargs):
        Service.__init__(self)
        dict.__init__(self)
        self.update(iterable, **kwargs)

    def __eq__(self, other):
        if not isinstance(other, PropertySet):
            other = PropertySet(other)
        return dict.__eq__(self, other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        x = 0
        for key, value in self.items():
            if isinstance(value, list):
                x ^= hash((key, tuple(value)))
            else:
                x ^= hash((key, value))
        return x

    def __getitem__(self, key):
        return dict.get(self, key)

    def __setitem__(self, key, value):
        if value is None:
            try:
                dict.__delitem__(self, key)
            except KeyError:
                pass
        else:
            dict.__setitem__(self, key, cast_property(value))

    def pull(self):
        """ Copy the set of remote properties onto the local set.
        """
        self.resource.get()
        properties = self.resource.metadata
        self.replace(properties or {})

    def push(self):
        """ Copy the set of local properties onto the remote set.
        """
        self.resource.put(self)

    def replace(self, iterable=None, **kwargs):
        self.clear()
        self.update(iterable, **kwargs)

    def setdefault(self, key, default=None):
        if key in self:
            value = self[key]
        elif default is None:
            value = None
        else:
            value = dict.setdefault(self, key, default)
        return value

    def update(self, iterable=None, **kwargs):
        if iterable:
            try:
                for key in iterable.keys():
                    self[key] = iterable[key]
            except (AttributeError, TypeError):
                for key, value in iterable:
                    self[key] = value
        for key in kwargs:
            self[key] = kwargs[key]


class LabelSet(Service, set):
    """ A set subclass that can be bound to a remote *labels* resource.
    """

    def __init__(self, iterable=None):
        Service.__init__(self)
        set.__init__(self)
        if iterable:
            self.update(iterable)

    def __eq__(self, other):
        if not isinstance(other, LabelSet):
            other = LabelSet(other)
        return set.__eq__(self, other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        value = 0
        for label in self:
            value ^= hash(label)
        return value

    def pull(self):
        """ Copy the set of remote labels onto the local set.
        """
        self.resource.get()
        labels = self.resource.metadata
        self.replace(labels or [])

    def push(self):
        """ Copy the set of local labels onto the remote set.
        """
        self.resource.put(self)

    def replace(self, iterable):
        self.clear()
        self.update(iterable)


class PropertyContainer(Service):
    """ Base class for objects that contain a set of properties,
    i.e. :py:class:`Node` and :py:class:`Relationship`.
    """

    def __init__(self, **properties):
        Service.__init__(self)
        self.__properties = PropertySet(properties)
        # Auto-sync will be removed in 2.1
        self.auto_sync_properties = Graph.auto_sync_properties

    def __eq__(self, other):
        return self.properties == other.properties

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__properties)

    def __contains__(self, key):
        self.__pull_if_bound()
        return key in self.properties

    def __getitem__(self, key):
        self.__pull_if_bound()
        return self.properties.__getitem__(key)

    def __setitem__(self, key, value):
        self.properties.__setitem__(key, value)
        self.__push_if_bound()

    def __delitem__(self, key):
        self.properties.__delitem__(key)
        self.__push_if_bound()

    def bind(self, uri, metadata=None):
        """ Associate this «class.lower» with a remote resource.

        :arg uri: The URI identifying the remote resource to which to bind.
        :arg metadata: Dictionary of initial metadata to attach to the contained resource.

        """
        Service.bind(self, uri, metadata)
        self.__properties.bind(uri + "/properties")

    @property
    def properties(self):
        """ The set of properties attached to this «class.lower». Properties
        can also be read from and written to any :class:`PropertyContainer`
        by using the index syntax directly.
        """
        return self.__properties

    def pull(self):
        """ Pull data to this «class.lower» from its remote counterpart.
        """
        self.resource.get()
        properties = self.resource.metadata["data"]
        self.__properties.replace(properties or {})

    def push(self):
        """ Push data from this «class.lower» to its remote counterpart.
        """
        self.__properties.push()

    def unbind(self):
        """ Detach this «class.lower» from any remote counterpart.
        """
        Service.unbind(self)
        self.__properties.unbind()

    def __pull_if_bound(self):
        # remove in 2.1
        if self.auto_sync_properties:
            try:
                self.properties.pull()
            except BindError:
                pass

    def __push_if_bound(self):
        # remove in 2.1
        if self.auto_sync_properties:
            try:
                self.properties.push()
            except BindError:
                pass


class Node(PropertyContainer):
    """ A graph node that may optionally be bound to a remote counterpart
    in a Neo4j database. Nodes may contain a set of named :attr:`~py2neo.Node.properties` and
    may have one or more :attr:`.labels` applied to them::

        >>> from py2neo import Node
        >>> alice = Node("Person", name="Alice")
        >>> banana = Node("Fruit", "Food", colour="yellow", tasty=True)

    All positional arguments passed to the constructor are interpreted
    as labels and all keyword arguments as properties. It is also possible to
    construct Node instances from other data types (such as a dictionary)
    by using the :meth:`.Node.cast` class method::

        >>> bob = Node.cast({"name": "Bob Robertson", "age": 44})

    Labels and properties can be accessed and modified using the
    :attr:`.labels` and :attr:`~py2neo.Node.properties` attributes respectively.
    The former is an instance of :class:`.LabelSet`, which extends the
    built-in :class:`set` class, and the latter is an instance of
    :class:`.PropertySet` which extends :class:`dict`.

        >>> alice.properties["name"]
        'Alice'
        >>> alice.labels
        {'Person'}
        >>> alice.labels.add("Employee")
        >>> alice.properties["employee_no"] = 3456
        >>> alice
        <Node labels={'Employee', 'Person'} properties={'employee_no': 3456, 'name': 'Alice'}>

    One of the core differences between a :class:`.PropertySet` and a standard
    dictionary is in how it handles :const:`None` and missing values. As with actual Neo4j
    properties, missing values and those equal to :const:`None` are equivalent.
    """

    cache = ThreadLocalWeakValueDictionary()

    __id = None

    @staticmethod
    def cast(*args, **kwargs):
        """ Cast the arguments provided to a :class:`.Node` (or
        :class:`.NodePointer`). The following combinations of
        arguments are possible::

            >>> Node.cast(None)
            >>> Node.cast()
            <Node labels=set() properties={}>
            >>> Node.cast("Person")
            <Node labels={'Person'} properties={}>
            >>> Node.cast(name="Alice")
            <Node labels=set() properties={'name': 'Alice'}>
            >>> Node.cast("Person", name="Alice")
            <Node labels={'Person'} properties={'name': 'Alice'}>
            >>> Node.cast(123)
            <NodePointer address=123>
            >>> Node.cast({"name": "Alice"})
            <Node labels=set() properties={'name': 'Alice'}>
            >>> node = Node("Person", name="Alice")
            >>> Node.cast(node)
            <Node labels={'Person'} properties={'name': 'Alice'}>

        """
        if len(args) == 1 and not kwargs:
            from py2neo.batch import Job
            arg = args[0]
            if arg is None:
                return None
            elif isinstance(arg, (Node, NodePointer, Job)):
                return arg
            elif is_integer(arg):
                return NodePointer(arg)

        inst = Node()

        def apply(x):
            if isinstance(x, dict):
                inst.properties.update(x)
            elif is_collection(x):
                for item in x:
                    apply(item)
            else:
                inst.labels.add(ustr(x))

        for arg in args:
            apply(arg)
        inst.properties.update(kwargs)
        return inst

    @classmethod
    def hydrate(cls, data, inst=None):
        """ Hydrate a dictionary of data to produce a :class:`.Node` instance.
        The data structure and values expected are those produced by the
        `REST API <http://neo4j.com/docs/stable/rest-api-nodes.html#rest-api-get-node>`__
        although only the ``self`` value is required.

        :arg data: dictionary of data to hydrate
        :arg inst: an existing :class:`.Node` instance to overwrite with new values

        """
        self = data["self"]
        if inst is None:
            new_inst = cls()
            new_inst.__stale.update({"labels", "properties"})
            inst = cls.cache.setdefault(self, new_inst)
        cls.cache[self] = inst
        inst.bind(self, data)
        if "data" in data:
            inst.__stale.discard("properties")
            properties = data["data"]
            properties.update(inst.properties)
            inst._PropertyContainer__properties.replace(properties)
        if "metadata" in data:
            inst.__stale.discard("labels")
            metadata = data["metadata"]
            labels = set(metadata["labels"])
            labels.update(inst.labels)
            inst.__labels.replace(labels)
        return inst

    @classmethod
    def __joinable(cls, obj):
        from py2neo.batch import Job
        return obj is None or isinstance(obj, (Node, NodePointer, Job))

    @classmethod
    def join(cls, n, m):
        # Attempt to join two nodes together as single node.
        if not cls.__joinable(n) or not cls.__joinable(m):
            raise TypeError("Can only join Node, NodePointer, Job or None")
        if n is None:
            return m
        elif m is None or n is m:
            return n
        elif isinstance(n, NodePointer) or isinstance(m, NodePointer):
            if isinstance(n, NodePointer) and isinstance(m, NodePointer) and n.address == m.address:
                return n
        elif n.bound and m.bound:
            if n.resource == m.resource:
                return n
        raise JoinError("Cannot join nodes {} and {}".format(n, m))

    def __init__(self, *labels, **properties):
        PropertyContainer.__init__(self, **properties)
        self.__labels = LabelSet(labels)
        self.__stale = set()

    def __repr__(self):
        s = [self.__class__.__name__]
        if self.bound:
            s.append("graph=%r" % self.graph.uri.string)
            s.append("ref=%r" % self.ref)
            if "labels" in self.__stale:
                s.append("labels=?")
            else:
                s.append("labels=%r" % set(self.labels))
            if "properties" in self.__stale:
                s.append("properties=?")
            else:
                s.append("properties=%r" % self.properties)
        else:
            s.append("labels=%r" % set(self.labels))
            s.append("properties=%r" % self.properties)
        return "<" + " ".join(s) + ">"

    def __str__(self):
        return xstr(self.__unicode__())

    def __unicode__(self):
        from py2neo.cypher import CypherWriter
        string = StringIO()
        writer = CypherWriter(string)
        if self.bound:
            writer.write_node(self, "n" + ustr(self._id))
        else:
            writer.write_node(self)
        return string.getvalue()

    def __eq__(self, other):
        if other is None:
            return False
        other = Node.cast(other)
        if self.bound and other.bound:
            return self.resource == other.resource
        else:
            return (LabelSet.__eq__(self.labels, other.labels) and
                    PropertyContainer.__eq__(self, other))

    def __hash__(self):
        value = super(Node, self).__hash__() ^ hash(self.labels)
        if self.bound:
            value ^= hash(self.resource.uri)
        return value

    def __add__(self, other):
        return Path(self, other)

    @property
    def _id(self):
        """ The internal ID of this node within the database.
        """
        if self.__id is None:
            self.__id = int(self.uri.path.segments[-1])
        return self.__id

    @property
    def ref(self):
        """ The URI of this node relative to its graph.

        :rtype: string
        """
        return "node/%s" % self._id

    def bind(self, uri, metadata=None):
        """ Associate this node with a remote node.

        :arg uri: The URI identifying the remote node to which to bind.
        :arg metadata: Dictionary of initial metadata to attach to the contained resource.

        """
        PropertyContainer.bind(self, uri, metadata)
        if self.graph.supports_node_labels:
            self.__labels.bind(uri + "/labels")
        else:
            from py2neo.legacy.core import LegacyNode
            self.__class__ = LegacyNode
        self.cache[uri] = self

    @property
    def degree(self):
        """ The number of relationships attached to this node.
        """
        statement = "START n=node({n}) MATCH (n)-[r]-() RETURN count(r)"
        return self.graph.cypher.execute_one(statement, {"n": self})

    @property
    def exists(self):
        """ :const:`True` if this node exists in the database,
        :const:`False` otherwise.
        """
        try:
            self.resource.get()
        except GraphError as error:
            if error.__cause__ and error.__cause__.status_code == NOT_FOUND:
                return False
            else:
                raise
        else:
            return True

    @property
    def labels(self):
        """ The set of labels attached to this node.
        """
        if self.bound and "labels" in self.__stale:
            self.refresh()
        return self.__labels

    def match(self, rel_type=None, other_node=None, limit=None):
        """ Return an iterator for all relationships attached to this node
        that match the specified criteria. See :meth:`.Graph.match` for
        argument details.
        """
        return self.graph.match(self, rel_type, other_node, True, limit)

    def match_incoming(self, rel_type=None, start_node=None, limit=None):
        """ Return an iterator for all incoming relationships to this node
        that match the specified criteria. See :meth:`.Graph.match` for
        argument details.
        """
        return self.graph.match(start_node, rel_type, self, False, limit)

    def match_outgoing(self, rel_type=None, end_node=None, limit=None):
        """ Return an iterator for all outgoing relationships from this node
        that match the specified criteria. See :meth:`.Graph.match` for
        argument details.
        """
        return self.graph.match(self, rel_type, end_node, False, limit)

    @property
    def properties(self):
        """ The set of properties attached to this node. Properties
        can also be read from and written to any :class:`Node`
        by using the index syntax directly. This means
        the following statements are equivalent::

            node.properties["name"] = "Alice"
            node["name"] = "Alice"

        """
        if self.bound and "properties" in self.__stale:
            self.refresh()
        return super(Node, self).properties

    def pull(self):
        """ Pull data to this node from its remote counterpart. Consider
        using :meth:`.Graph.pull` instead for batches of nodes.
        """
        super(Node, self).properties.clear()
        self.__labels.clear()
        self.refresh()

    def push(self):
        """ Push data from this node to its remote counterpart. Consider
        using :meth:`.Graph.push` instead for batches of nodes.
        """
        from py2neo.batch.push import PushBatch
        batch = PushBatch(self.graph)
        batch.append(self)
        batch.push()

    def refresh(self):
        # Non-destructive pull.
        query = "START a=node({a}) RETURN a,labels(a)"
        content = self.graph.cypher.post(query, {"a": self._id}).content
        dehydrated, label_metadata = content["data"][0]
        dehydrated.setdefault("metadata", {})["labels"] = label_metadata
        Node.hydrate(dehydrated, self)

    def unbind(self):
        """ Detach this node from any remote counterpart.
        """
        try:
            del self.cache[self.uri]
        except KeyError:
            pass
        PropertyContainer.unbind(self)
        self.__labels.unbind()
        self.__id = None


class NodePointer(object):
    """ Pointer to a :class:`Node` object. This can be used in a batch
    context to point to a node not yet created.
    """

    #: The address or index to which this pointer points.
    address = None

    def __init__(self, address):
        self.address = address

    def __repr__(self):
        return "<NodePointer address=%s>" % self.address

    def __str__(self):
        return xstr(self.__unicode__())

    def __unicode__(self):
        return "{%s}" % self.address

    def __eq__(self, other):
        return self.address == other.address

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.address)


class Rel(PropertyContainer):
    """ A :class:`.Rel` is similar to a :class:`.Relationship` but does not
    store information about the nodes to which it is attached. This class
    is used internally to bundle relationship type and property details
    for :class:`.Relationship` and :class:`.Path` objects but may also be
    used to denote an explicit forward relationship within a :class:`.Path`.

    .. seealso:: :class:`py2neo.Rev`

    """

    cache = ThreadLocalWeakValueDictionary()
    pair = None
    _pair_class = object

    __id = None

    @staticmethod
    def cast(*args, **kwargs):
        """ Cast the arguments provided to a :class:`.«class»`. The
        following combinations of arguments are possible::

            >>> «class».cast(None)
            >>> «class».cast()
            <«class» type=None properties={}>
            >>> «class».cast("KNOWS")
            <«class» type='KNOWS' properties={}>
            >>> «class».cast("KNOWS", since=1999)
            <«class» type='KNOWS' properties={'since': 1999}>
            >>> «class».cast("KNOWS", {"since": 1999})
            <«class» type='KNOWS' properties={'since': 1999}>
            >>> «class».cast(("KNOWS",))
            <«class» type='KNOWS' properties={}>
            >>> «class».cast(("KNOWS", {"since": 1999}))
            <«class» type='KNOWS' properties={'since': 1999}>
            >>> «class.lower» = «class»("KNOWS", since=1999)
            >>> «class».cast(«class.lower»)
            <«class» type='KNOWS' properties={'since': 1999}>

        """
        if len(args) == 1 and not kwargs:
            from py2neo.batch import Job
            arg = args[0]
            if arg is None:
                return None
            elif isinstance(arg, (Rel, Job)):
                return arg
            elif isinstance(arg, Relationship):
                return arg.rel

        inst = Rel()

        def apply(x):
            if isinstance(x, dict):
                inst.properties.update(x)
            elif is_collection(x):
                for item in x:
                    apply(item)
            else:
                inst.type = ustr(x)

        for arg in args:
            apply(arg)
        inst.properties.update(kwargs)
        return inst

    @classmethod
    def hydrate(cls, data, inst=None):
        """ Hydrate a dictionary of data to produce a :class:`.«class»` instance.
        The data structure and values expected are those produced by the
        `REST API <http://neo4j.com/docs/stable/rest-api-relationships.html#rest-api-get-relationship-by-id>`__
        although only the ``self`` value is required.

        :arg data: dictionary of data to hydrate
        :arg inst: an existing :class:`.«class»` instance to overwrite with new values

        """
        self = data["self"]
        if inst is None:
            new_inst = cls()
            new_inst.__stale.update({"properties"})
            inst = cls.cache.setdefault(self, new_inst)
        cls.cache[self] = inst
        inst.bind(self, data)
        inst.__type = data.get("type")
        pair = inst.pair
        if pair is not None:
            pair._Rel__type = inst.__type
        if "data" in data:
            inst.__stale.discard("properties")
            properties = data["data"]
            properties.update(inst.properties)
            inst._PropertyContainer__properties.replace(properties)
        return inst

    def __init__(self, *type_, **properties):
        if len(type_) > 1:
            raise ValueError("Only one relationship type can be specified")
        PropertyContainer.__init__(self, **properties)
        self.__type = type_[0] if type_ else None
        self.__stale = set()

    def __repr__(self):
        s = [self.__class__.__name__]
        if self.bound:
            s.append("graph=%r" % self.graph.uri.string)
            s.append("ref=%r" % self.ref)
            if self.__type is None:
                s.append("type=?")
            else:
                s.append("type=%r" % self.type)
            if "properties" in self.__stale:
                s.append("properties=?")
            else:
                s.append("properties=%r" % self.properties)
        else:
            s.append("type=%r" % self.type)
            s.append("properties=%r" % self.properties)
        return "<" + " ".join(s) + ">"

    def __str__(self):
        return xstr(self.__unicode__())

    def __unicode__(self):
        from py2neo.cypher import CypherWriter
        string = StringIO()
        writer = CypherWriter(string)
        if self.bound:
            writer.write_rel(self, "r" + ustr(self._id))
        else:
            writer.write_rel(self)
        return string.getvalue()

    def __eq__(self, other):
        if other is None:
            return False
        other = Rel.cast(other)
        if self.bound and other.bound:
            return self.resource == other.resource
        else:
            return self.type == other.type and self.properties == other.properties

    def __hash__(self):
        value = super(Rel, self).__hash__() ^ hash(self.type)
        if self.bound:
            value ^= hash(self.resource.uri)
        return value

    def __pos__(self):
        return self

    def __neg__(self):
        if self.pair is None:
            self.pair = self._pair_class()
            self.pair.__resource__ = self.__resource__
            self.pair._PropertyContainer__properties = self._PropertyContainer__properties
            self.pair._Rel__type = self.__type
            self.pair._Rel__stale = self.__stale
            self.pair.pair = self
        return self.pair

    def __abs__(self):
        return self

    @property
    def _id(self):
        """ The internal ID of this relationship within the database.
        """
        if self.__id is None:
            self.__id = int(self.uri.path.segments[-1])
        return self.__id

    @property
    def ref(self):
        """ The URI of this relationship relative to its graph.

        :rtype: string
        """
        return "relationship/%s" % self._id

    def bind(self, uri, metadata=None):
        """ Associate this object with a remote relationship.

        :arg uri: The URI identifying the remote relationship to which to bind.
        :arg metadata: Dictionary of initial metadata to attach to the contained resource.

        """
        PropertyContainer.bind(self, uri, metadata)
        self.cache[uri] = self
        pair = self.pair
        if pair is not None:
            PropertyContainer.bind(pair, uri, metadata)
            # make sure we're using exactly the same resource object
            # (maybe could write a Service.multi_bind classmethod
            pair.__resource__ = self.__resource__
            pair.cache[uri] = pair

    @property
    def exists(self):
        """ :const:`True` if this relationship exists in the database,
        :const:`False` otherwise.
        """
        try:
            self.resource.get()
        except GraphError as error:
            if error.__cause__ and error.__cause__.status_code == NOT_FOUND:
                return False
            else:
                raise
        else:
            return True

    @property
    def properties(self):
        """ The set of properties attached to this relationship. Properties
        can also be read from and written to any :class:`Rel`
        by using the index syntax directly. This means
        the following statements are equivalent::

            rel.properties["since"] = 1999
            rel["since"] = 1999

        """
        if self.bound and "properties" in self.__stale:
            self.refresh()
        return super(Rel, self).properties

    def pull(self):
        """ Pull data to this relationship from its remote counterpart.
        """
        super(Rel, self).properties.clear()
        self.refresh()

    def push(self):
        """ Push data from this relationship to its remote counterpart.
        """
        super(Rel, self).push()

    def refresh(self):
        # Non-destructive pull.
        super(Rel, self).pull()
        pulled_type = self.resource.metadata["type"]
        self.__type = pulled_type
        pair = self.pair
        if pair is not None:
            pair._Rel__type = pulled_type
        self.__stale.clear()

    @property
    def type(self):
        """ The type of this relationship.
        """
        if self.bound and self.__type is None:
            self.pull()
        return self.__type

    @type.setter
    def type(self, name):
        """ Set the type of this relationship (only possible if not bound).
        """
        if self.bound:
            raise AttributeError("The type of a bound Rel is immutable")
        self.__type = name
        pair = self.pair
        if pair is not None:
            pair._Rel__type = name

    def unbind(self):
        """ Detach this relationship from any remote counterpart.
        """
        try:
            del self.cache[self.uri]
        except KeyError:
            pass
        PropertyContainer.unbind(self)
        self.__id = None
        pair = self.pair
        if pair is not None:
            try:
                del pair.cache[pair.uri]
            except KeyError:
                pass
            PropertyContainer.unbind(pair)


class Rev(Rel):
    """ A :class:`.Rev` is identical to a :class:`.Rel` but denotes a
    reversed relationship rather than a forward one. The following
    example shows how to build a :class:`.Path` with one forward and
    one reversed relationship::

        >>> path = Path(Node(name="A"), Rel("TO"), Node(name="B"), Rev("TO"), Node(name="C"))
        >>> for relationship in path.relationships:
        ...     print(relationship)
        ({name:"A"})-[:TO]->({name:"B"})
        ({name:"C"})-[:TO]->({name:"B"})

    .. seealso:: :class:`py2neo.Rel`

    """

    _pair_class = Rel

    def __abs__(self):
        return self.__neg__()

    def __hash__(self):
        return -(super(Rev, self).__hash__())


Rel._pair_class = Rev


class Path(object):
    """ A sequence of nodes connected by relationships that may
    optionally be bound to remote counterparts in a Neo4j database.

        >>> from py2neo import Node, Path, Rev
        >>> alice, bob, carol = Node(name="Alice"), Node(name="Bob"), Node(name="Carol")
        >>> abc = Path(alice, "KNOWS", bob, Rev("KNOWS"), carol)
        >>> abc
        <Path order=3 size=2>
        >>> abc.nodes
        (<Node labels=set() properties={'name': 'Alice'}>,
         <Node labels=set() properties={'name': 'Bob'}>,
         <Node labels=set() properties={'name': 'Carol'}>)
        >>> abc.rels
        (<Rel type='KNOWS' properties={}>, <Rev type='KNOWS' properties={}>)
        >>> abc.relationships
        (<Relationship type='KNOWS' properties={}>,
         <Relationship type='KNOWS' properties={}>)
        >>> dave, eve = Node(name="Dave"), Node(name="Eve")
        >>> de = Path(dave, "KNOWS", eve)
        >>> de
        <Path order=2 size=1>
        >>> abcde = Path(abc, "KNOWS", de)
        >>> abcde
        <Path order=5 size=4>
        >>> for relationship in abcde.relationships:
        ...     print(relationship)
        ({name:"Alice"})-[:KNOWS]->({name:"Bob"})
        ({name:"Carol"})-[:KNOWS]->({name:"Bob"})
        ({name:"Carol"})-[:KNOWS]->({name:"Dave"})
        ({name:"Dave"})-[:KNOWS]->({name:"Eve"})

    """

    @classmethod
    def hydrate(cls, data, inst=None):
        """ Hydrate a dictionary of data to produce a :class:`.Path` instance.
        The data structure and values expected are those produced by the
        `REST API <http://neo4j.com/docs/stable/rest-api-graph-algos.html#rest-api-find-one-of-the-shortest-paths>`__.

        :arg data: dictionary of data to hydrate
        :arg inst: an existing :class:`.Path` instance to overwrite with new values

        """
        node_uris = data["nodes"]
        relationship_uris = data["relationships"]
        rel_rev = [Rel if direction == "->" else Rev for direction in data["directions"]]
        if inst is None:
            nodes = [Node.hydrate({"self": uri}) for uri in node_uris]
            rels = [rel_rev[i].hydrate({"self": uri}) for i, uri in enumerate(relationship_uris)]
            inst = Path(*round_robin(nodes, rels))
        else:
            for i, node in enumerate(inst.nodes):
                uri = node_uris[i]
                Node.hydrate({"self": uri}, node)
            for i, rel in enumerate(inst.rels):
                uri = relationship_uris[i]
                rel_rev[i].hydrate({"self": uri}, rel)
        inst.__metadata = data
        return inst

    def __init__(self, *entities):
        nodes = []
        rels = []

        def join_path(path, index):
            if len(nodes) == len(rels):
                nodes.extend(path.nodes)
                rels.extend(path.rels)
            else:
                # try joining forward
                try:
                    nodes[-1] = Node.join(nodes[-1], path.start_node)
                except JoinError:
                    # try joining backward
                    try:
                        nodes[-1] = Node.join(nodes[-1], path.end_node)
                    except JoinError:
                        raise JoinError("Path at position %s cannot be joined" % index)
                    else:
                        nodes.extend(path.nodes[-2::-1])
                        rels.extend(-r for r in path.rels[::-1])
                else:
                    nodes.extend(path.nodes[1:])
                    rels.extend(path.rels)

        def join_rel(rel, index):
            if len(nodes) == len(rels):
                raise JoinError("Rel at position %s cannot be joined" % index)
            else:
                rels.append(rel)

        def join_node(node):
            if len(nodes) == len(rels):
                nodes.append(node)
            else:
                nodes[-1] = Node.join(nodes[-1], node)

        for i, entity in enumerate(entities):
            if isinstance(entity, Path):
                join_path(entity, i)
            elif isinstance(entity, Rel):
                join_rel(entity, i)
            elif isinstance(entity, (Node, NodePointer)):
                join_node(entity)
            elif len(nodes) == len(rels):
                join_node(Node.cast(entity))
            else:
                join_rel(Rel.cast(entity), i)
        join_node(None)

        self.__nodes = tuple(nodes)
        self.__rels = tuple(rels)
        self.__relationships = None
        self.__order = len(self.__nodes)
        self.__size = len(self.__rels)
        self.__metadata = None

    def __repr__(self):
        s = [self.__class__.__name__]
        if self.bound:
            s.append("graph=%r" % self.graph.uri.string)
            s.append("start=%r" % self.start_node.ref)
            s.append("end=%r" % self.end_node.ref)
        s.append("order=%r" % self.order)
        s.append("size=%r" % self.size)
        return "<" + " ".join(s) + ">"

    def __str__(self):
        return xstr(self.__unicode__())

    def __unicode__(self):
        from py2neo.cypher import CypherWriter
        string = StringIO()
        writer = CypherWriter(string)
        writer.write_path(self)
        return string.getvalue()

    def __eq__(self, other):
        try:
            return self.nodes == other.nodes and self.rels == other.rels
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        value = 0
        for entity in self.rels + self.nodes:
            value ^= hash(entity)
        return value

    def __bool__(self):
        return bool(self.rels)

    def __nonzero__(self):
        return bool(self.rels)

    def __len__(self):
        return self.size

    def __getitem__(self, item):
        try:
            if isinstance(item, slice):
                path = Path()
                p, q = item.start, item.stop
                if q is not None:
                    q += 1
                path.__nodes = self.nodes[p:q]
                path.__rels = self.rels[item]
                return path
            else:
                if item >= 0:
                    start_node = self.nodes[item]
                    end_node = self.nodes[item + 1]
                else:
                    start_node = self.nodes[item - 1]
                    end_node = self.nodes[item]
                return Relationship(start_node, self.rels[item], end_node)
        except IndexError:
            raise IndexError("Path segment index out of range")

    def __iter__(self):
        return iter(self.relationships)

    def __reversed__(self):
        return iter(reversed(self.relationships))

    def __add__(self, other):
        return Path(self, other)

    @property
    def bound(self):
        """ :const:`True` if this path is bound to a remote counterpart,
        :const:`False` otherwise.
        """
        try:
            _ = self.service_root
        except BindError:
            return False
        else:
            return True

    @property
    def end_node(self):
        """ The end node of this «class.lower».

        :return: :class:`.Node`
        """
        return self.__nodes[-1]

    @property
    def exists(self):
        """ :const:`True` if this path exists in the database,
        :const:`False` otherwise.
        """
        return all(entity.exists for entity in self.nodes + self.rels)

    @property
    def graph(self):
        """ The parent graph of this path.

        :rtype: :class:`.Graph`
        """
        return self.service_root.graph

    @property
    def nodes(self):
        """ A tuple of all nodes in this «class.lower».
        """
        return self.__nodes

    @property
    def order(self):
        """ The number of nodes in this «class.lower».
        """
        return self.__order

    def pull(self):
        """ Pull data to all entities in this path from their remote counterparts.
        """
        from py2neo.batch.pull import PullBatch
        batch = PullBatch(self.graph)
        for relationship in self:
            batch.append(relationship)
        batch.pull()

    def push(self):
        """ Push data from all entities in this path to their remote counterparts.
        """
        from py2neo.batch.push import PushBatch
        batch = PushBatch(self.graph)
        for relationship in self:
            batch.append(relationship)
        batch.push()

    @property
    def relationships(self):
        """ A tuple of all relationships in this «class.lower».
        """
        if self.__relationships is None:
            self.__relationships = tuple(
                Relationship(self.nodes[i], rel, self.nodes[i + 1])
                for i, rel in enumerate(self.rels)
            )
        return self.__relationships

    @property
    def rels(self):
        """ A tuple of all rels in this «class.lower».
        """
        return self.__rels

    @property
    def service_root(self):
        """ The root service associated with this path.

        :return: :class:`.ServiceRoot`
        """
        for relationship in self:
            try:
                return relationship.service_root
            except BindError:
                pass
        raise BindError("Local path is not bound to a remote path")

    @property
    def size(self):
        """ The number of relationships in this «class.lower».
        """
        return self.__size

    @property
    def start_node(self):
        """ The start node of this «class.lower».

        :return: :class:`.Node`
        """
        return self.__nodes[0]

    def unbind(self):
        """ Detach all entities in this path
        from any remote counterparts.
        """
        for entity in self.rels + self.nodes:
            try:
                entity.unbind()
            except BindError:
                pass


class Relationship(Path):
    """ A graph relationship that may optionally be bound to a remote counterpart
    in a Neo4j database. Relationships require a type name and may contain a set
    of named :attr:`~py2neo.Node.properties`.

    .. seealso::
       :class:`py2neo.Rel`
       :class:`py2neo.Rev`

    """

    cache = ThreadLocalWeakValueDictionary()

    __id = None

    @staticmethod
    def cast(*args, **kwargs):
        """ Cast the arguments provided to a :class:`.Relationship`. The
        following combinations of arguments are possible::

            >>> Relationship.cast(Node(), "KNOWS", Node())
            <Relationship type='KNOWS' properties={}>
            >>> Relationship.cast((Node(), "KNOWS", Node()))
            <Relationship type='KNOWS' properties={}>
            >>> Relationship.cast(Node(), "KNOWS", Node(), since=1999)
            <Relationship type='KNOWS' properties={'since': 1999}>
            >>> Relationship.cast(Node(), "KNOWS", Node(), {"since": 1999})
            <Relationship type='KNOWS' properties={'since': 1999}>
            >>> Relationship.cast((Node(), "KNOWS", Node(), {"since": 1999}))
            <Relationship type='KNOWS' properties={'since': 1999}>
            >>> Relationship.cast(Node(), ("KNOWS", {"since": 1999}), Node())
            <Relationship type='KNOWS' properties={'since': 1999}>
            >>> Relationship.cast((Node(), ("KNOWS", {"since": 1999}), Node()))
            <Relationship type='KNOWS' properties={'since': 1999}>
            >>> Relationship.cast(Node(), Rel("KNOWS", since=1999), Node())
            <Relationship type='KNOWS' properties={'since': 1999}>
            >>> Relationship.cast((Node(), Rel("KNOWS", since=1999), Node()))
            <Relationship type='KNOWS' properties={'since': 1999}>

        """
        if len(args) == 1 and not kwargs:
            arg = args[0]
            if isinstance(arg, Relationship):
                return arg
            elif isinstance(arg, tuple):
                if len(arg) == 3:
                    return Relationship(*arg)
                elif len(arg) == 4:
                    return Relationship(arg[0], arg[1], arg[2], **arg[3])
                else:
                    raise TypeError("Cannot cast relationship from {0}".format(arg))
            else:
                raise TypeError("Cannot cast relationship from {0}".format(arg))
        elif len(args) == 3:
            rel = Relationship(*args)
            rel.properties.update(kwargs)
            return rel
        elif len(args) == 4:
            props = args[3]
            props.update(kwargs)
            return Relationship(*args[0:3], **props)
        else:
            raise TypeError("Cannot cast relationship from {0}".format((args, kwargs)))

    @classmethod
    def hydrate(cls, data, inst=None):
        """ Hydrate a dictionary of data to produce a :class:`.Relationship` instance.
        The data structure and values expected are those produced by the
        `REST API <http://neo4j.com/docs/stable/rest-api-relationships.html#rest-api-get-relationship-by-id>`__.

        :arg data: dictionary of data to hydrate
        :arg inst: an existing :class:`.Relationship` instance to overwrite with new values

        """
        self = data["self"]
        if inst is None:
            inst = cls.cache.setdefault(self, cls(Node.hydrate({"self": data["start"]}),
                                                  Rel.hydrate(data),
                                                  Node.hydrate({"self": data["end"]})))
        else:
            Node.hydrate({"self": data["start"]}, inst.start_node)
            Node.hydrate({"self": data["end"]}, inst.end_node)
            Rel.hydrate(data, inst.rel)
        cls.cache[self] = inst
        return inst

    def __init__(self, start_node, rel, end_node, **properties):
        cast_rel = Rel.cast(rel)
        cast_rel._PropertyContainer__properties.update(properties)
        if isinstance(cast_rel, Rev):  # always forwards
            Path.__init__(self, end_node, -cast_rel, start_node)
        else:
            Path.__init__(self, start_node, cast_rel, end_node)

    def __repr__(self):
        s = [self.__class__.__name__]
        if self.bound:
            s.append("graph=%r" % self.graph.uri.string)
            s.append("ref=%r" % self.ref)
            s.append("start=%r" % self.start_node.ref)
            s.append("end=%r" % self.end_node.ref)
            if self.rel._Rel__type is None:
                s.append("type=?")
            else:
                s.append("type=%r" % self.type)
            if "properties" in self.rel._Rel__stale:
                s.append("properties=?")
            else:
                s.append("properties=%r" % self.properties)
        else:
            s.append("type=%r" % self.type)
            s.append("properties=%r" % self.properties)
        return "<" + " ".join(s) + ">"

    def __str__(self):
        return xstr(self.__unicode__())

    def __unicode__(self):
        from py2neo.cypher import CypherWriter
        string = StringIO()
        writer = CypherWriter(string)
        if self.bound:
            writer.write_relationship(self, "r" + ustr(self._id))
        else:
            writer.write_relationship(self)
        return string.getvalue()

    def __len__(self):
        return 1

    def __contains__(self, key):
        return self.rel.__contains__(key)

    def __getitem__(self, key):
        return self.rel.__getitem__(key)

    def __setitem__(self, key, value):
        self.rel.__setitem__(key, value)

    def __delitem__(self, key):
        self.rel.__delitem__(key)

    @property
    def _id(self):
        """ The internal ID of this relationship within the database.
        """
        if self.__id is None:
            self.__id = self.rel._id
        return self.__id

    def bind(self, uri, metadata=None):
        """ Associate this relationship with a remote relationship. The start and
        end nodes will also be associated with their corresponding remote nodes.

        :arg uri: The URI identifying the remote relationship to which to bind.
        :arg metadata: Dictionary of initial metadata to attach to the contained resource.

        """
        self.rel.bind(uri, metadata)
        self.cache[uri] = self
        for i, key, node in [(0, "start", self.start_node), (-1, "end", self.end_node)]:
            uri = self.resource.metadata[key]
            if isinstance(node, Node):
                node.bind(uri)
            else:
                nodes = list(self._Path__nodes)
                node = Node.cache.setdefault(uri, Node())
                if not node.bound:
                    node.bind(uri)
                nodes[i] = node
                self._Path__nodes = tuple(nodes)

    @property
    def bound(self):
        """ :const:`True` if this relationship is bound to a remote counterpart,
        :const:`False` otherwise.
        """
        return self.rel.bound

    @property
    def exists(self):
        """ :const:`True` if this relationship exists in the database,
        :const:`False` otherwise.
        """
        return self.rel.exists

    @property
    def graph(self):
        """ The parent graph of this relationship.

        :rtype: :class:`.Graph`
        """
        return self.service_root.graph

    @property
    def properties(self):
        """ The set of properties attached to this relationship. Properties
        can also be read from and written to any :class:`Relationship`
        by using the index syntax directly. This means
        the following statements are equivalent::

            relationship.properties["since"] = 1999
            relationship["since"] = 1999

        """
        return self.rel.properties

    def pull(self):
        """ Pull data to this relationship from its remote counterpart.
        """
        self.rel.pull()

    def push(self):
        """ Push data from this relationship to its remote counterpart.
        """
        self.rel.push()

    @property
    def ref(self):
        """ The URI of this relationship relative to its graph.

        :rtype: string
        """
        return self.rel.ref

    @property
    def rel(self):
        """ The :class:`.Rel` object within this relationship.
        """
        return self.rels[0]

    @property
    def resource(self):
        """ The resource object wrapped by this relationship, if
        bound.
        """
        return self.rel.resource

    @property
    def service_root(self):
        """ The root service associated with this relationship.

        :return: :class:`.ServiceRoot`
        """
        try:
            return self.rel.service_root
        except BindError:
            try:
                return self.start_node.service_root
            except BindError:
                return self.end_node.service_root

    @property
    def size(self):
        """ The number of relationships in this relationship. This property
        always equals 1 for a :class:`.Relationship` and is inherited from
        the more general parent class, :class:`.Path`.
        """
        return super(Relationship, self).size

    @property
    def type(self):
        """ The type of this relationship.
        """
        return self.rel.type

    @type.setter
    def type(self, name):
        """ Set the type of this relationship (only possible if not bound).
        """
        if self.rel.bound:
            raise AttributeError("The type of a bound Relationship is immutable")
        self.rel.type = name

    def unbind(self):
        """ Detach this relationship and its start and end
        nodes from any remote counterparts.
        """
        try:
            del self.cache[self.uri]
        except KeyError:
            pass
        self.rel.unbind()
        for node in [self.start_node, self.end_node]:
            if isinstance(node, Node):
                try:
                    node.unbind()
                except BindError:
                    pass
        self.__id = None

    @property
    def uri(self):
        """ The URI of this relationship, if bound.
        """
        return self.rel.uri


class Subgraph(object):
    """ A collection of :class:`Node` and :class:`Relationship` objects.
    """

    def __init__(self, *entities):
        self.__nodes = set()
        self.__relationships = set()
        for entity in entities:
            entity = Graph.cast(entity)
            if isinstance(entity, Node):
                self.__nodes.add(entity)
            elif isinstance(entity, Relationship):
                self.__nodes.add(entity.start_node)
                self.__nodes.add(entity.end_node)
                self.__relationships.add(entity)
            else:
                for node in entity.nodes:
                    self.__nodes.add(node)
                for relationship in entity.relationships:
                    self.__relationships.add(relationship)

    def __repr__(self):
        return "<Subgraph order=%s size=%s>" % (self.order, self.size)

    def __eq__(self, other):
        try:
            return self.nodes == other.nodes and self.relationships == other.relationships
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        value = 0
        for entity in self.__nodes | self.__relationships:
            value ^= hash(entity)
        return value

    def __bool__(self):
        return bool(self.__relationships)

    def __nonzero__(self):
        return bool(self.__relationships)

    def __len__(self):
        return self.size

    def __iter__(self):
        return iter(self.__relationships)

    def __contains__(self, item):
        if isinstance(item, Node):
            return item in self.__nodes
        elif isinstance(item, Path):
            for relationship in item:
                if relationship not in self.__relationships:
                    return False
            return True
        else:
            return False

    @property
    def bound(self):
        """ :const:`True` if all entities in this subgraph are bound to remote counterparts,
        :const:`False` otherwise.
        """
        try:
            _ = self.service_root
        except BindError:
            return False
        else:
            return True

    @property
    def exists(self):
        """ :const:`True` if this subgraph exists in the database,
        :const:`False` otherwise.
        """
        return all(entity.exists for entity in self.__nodes | self.__relationships)

    @property
    def graph(self):
        """ The parent graph of this subgraph.

        :rtype: :class:`.Graph`
        """
        return self.service_root.graph

    @property
    def nodes(self):
        """ The set of all nodes in this subgraph.
        """
        return frozenset(self.__nodes)

    @property
    def order(self):
        """ The number of nodes in this subgraph.
        """
        return len(self.__nodes)

    @property
    def relationships(self):
        """ The set of all relationships in this subgraph.
        """
        return frozenset(self.__relationships)

    @property
    def service_root(self):
        """ The root service associated with this subgraph.

        :return: :class:`.ServiceRoot`
        """
        for relationship in self:
            try:
                return relationship.service_root
            except BindError:
                pass
        raise BindError("Local path is not bound to a remote path")

    @property
    def size(self):
        """ The number of relationships in this subgraph.
        """
        return len(self.__relationships)

    def unbind(self):
        """ Detach all entities in this subgraph
        from any remote counterparts.
        """
        for entity in self.__nodes | self.__relationships:
            try:
                entity.unbind()
            except BindError:
                pass


class ServerPlugin(object):
    """ Base class for server plugins.
    """

    def __init__(self, graph, name):
        self.graph = graph
        self.name = name
        extensions = self.graph.resource.metadata["extensions"]
        try:
            self.resources = {key: Resource(value) for key, value in extensions[self.name].items()}
        except KeyError:
            raise LookupError("No plugin named %r found on graph <%s>" % (self.name, graph.uri))


class UnmanagedExtension(object):
    """ Base class for unmanaged extensions.
    """

    def __init__(self, graph, path):
        self.graph = graph
        self.resource = Resource(graph.service_root.uri.resolve(path))
        try:
            self.resource.get()
        except GraphError:
            raise NotImplementedError("No extension found at path %r on "
                                      "graph <%s>" % (path, graph.uri))


from py2neo.deprecated import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility module
"""


from __future__ import unicode_literals

try:
    from configparser import SafeConfigParser
except ImportError:
    from ConfigParser import SafeConfigParser
from itertools import cycle, islice
import os
import re
import sys
from threading import local
import warnings
from weakref import WeakValueDictionary


__all__ = ["numberise", "compact", "flatten", "round_robin", "deprecated",
           "version_tuple", "is_collection", "has_all", "pendulate",
           "is_integer", "is_string", "is_numeric", "ustr", "xstr",
           "PropertiesParser", "ThreadLocalWeakValueDictionary"]


def numberise(n):
    """ Convert a value to an integer if possible. If not, simply return
        the input value.
    """
    if n == "NaN":
        return None
    try:
        return int(n)
    except ValueError:
        return n


def compact(obj):
    """ Return a copy of an object with all :py:const:`None` values removed.
    """
    if isinstance(obj, dict):
        return dict((key, value) for key, value in obj.items() if value is not None)
    else:
        return obj.__class__(value for value in obj if value is not None)


def flatten(*values):
    for value in values:
        if hasattr(value, "__iter__"):
            for val in value:
                yield val
        else:
            yield value


def round_robin(*iterables):
    """ Cycle through a number of iterables, returning
        the next item from each in turn.

        round_robin('ABC', 'D', 'EF') --> A D E B F C

        Original recipe credited to George Sakkis
        Python 2/3 cross-compatibility tweak by Nigel Small
    """
    pending = len(iterables)
    nexts = cycle(iter(it) for it in iterables)
    while pending:
        try:
            for n in nexts:
                yield next(n)
        except StopIteration:
            pending -= 1
            nexts = cycle(islice(nexts, pending))


def deprecated(message):
    """ Decorator for deprecating functions and methods.

    ::

        @deprecated("'foo' has been deprecated in favour of 'bar'")
        def foo(x):
            pass

    """
    def f__(f):
        def f_(*args, **kwargs):
            warnings.warn(message, category=DeprecationWarning, stacklevel=2)
            return f(*args, **kwargs)
        f_.__name__ = f.__name__
        f_.__doc__ = f.__doc__
        f_.__dict__.update(f.__dict__)
        return f_
    return f__


VERSION = re.compile("(\d+\.\d+(\.\d+)?)")


def version_tuple(string):
    numbers = VERSION.match(string)
    if numbers:
        version = [int(n) for n in numbers.group(0).split(".")]
        extra = string[len(numbers.group(0)):]
        while extra.startswith(".") or extra.startswith("-"):
            extra = extra[1:]
    else:
        version = []
        extra = string
    while len(version) < 3:
        version += [0]
    if extra:
        version += [extra]
    return tuple(version)


def is_collection(obj):
    """ Returns true for any iterable which is not a string or byte sequence.
    """
    try:
        if isinstance(obj, unicode):
            return False
    except NameError:
        pass
    if isinstance(obj, bytes):
        return False
    try:
        iter(obj)
    except TypeError:
        return False
    try:
        hasattr(None, obj)
    except TypeError:
        return True
    return False


has_all = lambda iterable, items: all(item in iterable for item in items)


if sys.version_info >= (3,):

    is_integer = lambda x: isinstance(x, int)
    is_numeric = lambda x: isinstance(x, (int, float, complex))
    is_string = lambda x: isinstance(x, str)

    def ustr(s, encoding="utf-8"):
        """ Convert argument to unicode string.
        """
        if isinstance(s, str):
            return s
        try:
            return s.decode(encoding)
        except AttributeError:
            return str(s)

    def xstr(s, encoding="utf-8"):
        """ Convert argument to string type returned by __str__.
        """
        return ustr(s, encoding)

else:

    is_integer = lambda x: isinstance(x, (int, long))
    is_numeric = lambda x: isinstance(x, (int, float, long, complex))
    is_string = lambda x: isinstance(x, (str, unicode))

    def ustr(s, encoding="utf-8"):
        """ Convert argument to unicode string.
        """
        if isinstance(s, str):
            return s.decode(encoding)
        else:
            return unicode(s)

    def xstr(s, encoding="utf-8"):
        """ Convert argument to string type returned by __str__.
        """
        if isinstance(s, str):
            return s
        else:
            return unicode(s).encode(encoding)


def pendulate(collection):
    count = len(collection)
    for i in range(count):
        if i % 2 == 0:
            index = i // 2
        else:
            index = count - ((i + 1) // 2)
        yield index, collection[index]


def raise_from(exception, cause):
    exception.__cause__ = cause
    raise exception


if sys.version_info >= (3,):

    class PropertiesParser(SafeConfigParser):

        def read_properties(self, filename, section=None):
            if not section:
                basename = os.path.basename(filename)
                if basename.endswith(".properties"):
                    section = basename[:-11]
                else:
                    section = basename
            with open(filename) as f:
                data = f.read()
            self.read_string("[%s]\n%s" % (section, data), filename)

else:

    import codecs
    from io import StringIO

    class PropertiesParser(SafeConfigParser):

        def read_properties(self, filename, section=None):
            if not section:
                basename = os.path.basename(filename)
                if basename.endswith(".properties"):
                    section = basename[:-11]
                else:
                    section = basename
            data = StringIO()
            data.write("[%s]\n" % section)
            with codecs.open(filename, encoding="utf-8") as f:
                data.write(f.read())
            data.seek(0, os.SEEK_SET)
            self.readfp(data)


class ThreadLocalWeakValueDictionary(WeakValueDictionary, local):
    pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from subprocess import check_output
from shutil import rmtree

from py2neo.env import DIST_HOST
from py2neo.packages.httpstream import NetworkAddressError
from py2neo.server import download, dist_archive_name, GraphServer


__all__ = ["Warehouse", "Box"]


class Warehouse(object):
    """ Container for Neo4j :class:`.Box` installations.

    :arg home: Home directory for this warehouse (defaults to the path
               held in ``$NEOBOX_HOME`` if specified, otherwise ``$HOME/.neobox``)

    """

    __instances = {}

    def __new__(cls, home=None):
        if not home:
            home = os.getenv("NEOBOX_HOME") or os.path.expanduser("~/.neobox")
        key = home
        try:
            inst = cls.__instances[key]
        except KeyError:
            inst = super(Warehouse, cls).__new__(cls)
            inst.home = home
            cls.__instances[key] = inst
        return inst

    home = None

    def __repr__(self):
        return "<Warehouse home=%r>" % self.home

    def __hash__(self):
        return hash(self.home)

    def box(self, name):
        """ Fetch a named box. This box may or may not contain a Neo4j
        server installation.

        :arg name: Name of the box to fetch.
        :rtype: :class:`.Box`

        """
        return Box(self, name)

    def boxes(self):
        """ Fetch a list of all installed boxes.

        :rtype: generator

        """
        for name in os.listdir(os.path.join(self.home, "boxes")):
            yield Box(self, name)

    @property
    def _ports_path(self):
        return os.path.join(self.home, "ports")

    def ensure_downloaded(self, edition, version):
        dist_path = os.path.join(self.home, "dist")
        try:
            os.makedirs(dist_path)
        except FileExistsError:
            pass
        filename = os.path.join(dist_path, dist_archive_name(edition, version))
        if os.path.isfile(filename):
            return filename
        try:
            return download(edition, version, dist_path)
        except NetworkAddressError:
            raise RuntimeError("Not able to connect to %s" % DIST_HOST)

    @property
    def _ports(self):
        ports_path = self._ports_path
        try:
            os.makedirs(ports_path)
        except FileExistsError:
            pass
        ports = [port for port in os.listdir(ports_path)]
        return {os.path.basename(os.readlink(os.path.join(ports_path, port))): int(port)
                for port in ports}

    def _assign_port(self, name, port=None):
        if not port:
            ports = self._ports.values()
            if ports:
                port = max(ports) + 2
            else:
                port = 47470
        os.symlink(os.path.join("..", "boxes", name), os.path.join(self._ports_path, str(port)))
        return port

    def _remove_port(self, name):
        try:
            port = self._ports[name]
        except KeyError:
            pass
        else:
            os.remove(os.path.join(self._ports_path, str(port)))


class Box(object):
    """ Named container for a single Neo4j installation.

    :arg warehouse: The warehouse to which this box belongs.
    :arg name: The name of this box.

    """

    __instances = {}

    def __new__(cls, warehouse, name):
        key = (warehouse, name)
        try:
            inst = cls.__instances[key]
        except KeyError:
            inst = super(Box, cls).__new__(cls)
            inst.warehouse = warehouse
            inst.name = name
            cls.__instances[key] = inst
        return inst

    warehouse = None
    name = None

    def __repr__(self):
        return "<Box warehouse=%r name=%r>" % (self.warehouse.home, self.name)

    def __hash__(self):
        return hash(self.warehouse) ^ hash(self.name)

    @property
    def home(self):
        """ Base directory for this box. This directory may not exist.
        """
        return os.path.join(self.warehouse.home, "boxes", self.name)

    @property
    def server(self):
        """ The :class:`py2neo.server.GraphServer` installed in this box or
        :const:`None` if this box is empty.
        """
        if os.path.exists(self.home):
            return GraphServer(os.path.join(self.home, "neo4j"))
        else:
            return None

    def install(self, edition, version):
        """ Install a Neo4j server in this box.

        :arg edition: Neo4j edition (`'community'` or `'enterprise'`)
        :arg version: Neo4j version (e.g. `'2.1.5'`)
        :rtype: :class:`py2neo.server.GraphServer`

        """
        inst_path = self.home
        if self.server is not None:
            raise ValueError("A box named %r already exists" % self.name)
        filename = self.warehouse.ensure_downloaded(edition, version)
        os.makedirs(inst_path)
        # The Python tarfile module doesn't seem to recognise the Neo4j tar format :-(
        check_output("tar -x -C \"%s\" -f \"%s\"" % (inst_path, filename), shell=True)
        os.symlink(os.listdir(inst_path)[0], self.server.home)
        port = self.warehouse._assign_port(self.name)
        server = self.server
        server.update_server_properties(webserver_port=port, webserver_https_port=(port + 1))
        return server

    def remove(self, force=False):
        """ Remove any installed Neo4j server from this box. Running
        servers cannot be removed unless forced.

        :arg force: :const:`True` to remove server even if running.

        """
        if self.server.pid and not force:
            raise RuntimeError("Cannot remove a running server instance")
        self.warehouse._remove_port(self.name)
        rmtree(self.home, ignore_errors=(not force))

    def rename(self, new_name):
        """ Rename this box.

        :arg new_name: The new name for this box.

        """
        inst_path = self.home
        if not os.path.isdir(inst_path):
            raise ValueError("No box named %r exists" % self.name)
        new_box = self.warehouse.box(new_name)
        new_inst_path = new_box.home
        if os.path.isdir(new_inst_path):
            raise ValueError("A server instance named %r already exists" % new_name)
        port = self.warehouse._ports[self.name]
        self.warehouse._remove_port(self.name)
        os.rename(inst_path, new_inst_path)
        self.warehouse._assign_port(new_name, port)
        self.name = new_name



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2013-2014, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
JSONStream
"""


__all__ = ["__author__", "__copyright__", "__email__", "__license__",
           "__version__", "AwaitingData", "EndOfStream", "UnexpectedCharacter",
           "JSONStream", "assembled", "grouped"]

__author__ = "Nigel Small"
__copyright__ = "2013, Nigel Small"
__email__ = "nigel@nigelsmall.com"
__license__ = "Apache License, Version 2.0"
__version__ = "0.9.1"


import sys as __sys


if __sys.version_info >= (3,):
    try:
        from .cjsonstream import (
            AwaitingData, EndOfStream, UnexpectedCharacter, TextStream,
            Tokeniser, JSONStream, assembled, grouped
        )
    except ImportError:
        from .pyjsonstream import (
            AwaitingData, EndOfStream, UnexpectedCharacter, TextStream,
            Tokeniser, JSONStream, assembled, grouped
        )
else:
    try:
        from .cjsonstream_2x import (
            AwaitingData, EndOfStream, UnexpectedCharacter, TextStream,
            Tokeniser, JSONStream, assembled, grouped
        )
    except ImportError:
        from .pyjsonstream import (
            AwaitingData, EndOfStream, UnexpectedCharacter, TextStream,
            Tokeniser, JSONStream, assembled, grouped
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Bits and pieces copied from Python 3.3 source code for use with earlier versions.


from datetime import datetime, timedelta, tzinfo


class timezone(tzinfo):
    __slots__ = '_offset', '_name'

    # Sentinel value to disallow None
    _Omitted = object()
    def __new__(cls, offset, name=_Omitted):
        if not isinstance(offset, timedelta):
            raise TypeError("offset must be a timedelta")
        if name is cls._Omitted:
            if not offset:
                return cls.utc
            name = None
        elif not isinstance(name, str):
            raise TypeError("name must be a string")
        if not cls._minoffset <= offset <= cls._maxoffset:
            raise ValueError("offset must be a timedelta"
                             " strictly between -timedelta(hours=24) and"
                             " timedelta(hours=24).")
        if (offset.microseconds != 0 or
            offset.seconds % 60 != 0):
            raise ValueError("offset must be a timedelta"
                             " representing a whole number of minutes")
        return cls._create(offset, name)

    @classmethod
    def _create(cls, offset, name=None):
        self = tzinfo.__new__(cls)
        self._offset = offset
        self._name = name
        return self

    def __getinitargs__(self):
        """pickle support"""
        if self._name is None:
            return (self._offset,)
        return (self._offset, self._name)

    def __eq__(self, other):
        if type(other) != timezone:
            return False
        return self._offset == other._offset

    def __hash__(self):
        return hash(self._offset)

    def __repr__(self):
        """Convert to formal string, for repr().

        >>> tz = timezone.utc
        >>> repr(tz)
        'datetime.timezone.utc'
        >>> tz = timezone(timedelta(hours=-5), 'EST')
        >>> repr(tz)
        "datetime.timezone(datetime.timedelta(-1, 68400), 'EST')"
        """
        if self is self.utc:
            return 'datetime.timezone.utc'
        if self._name is None:
            return "%s(%r)" % ('datetime.' + self.__class__.__name__,
                               self._offset)
        return "%s(%r, %r)" % ('datetime.' + self.__class__.__name__,
                               self._offset, self._name)

    def __str__(self):
        return self.tzname(None)

    def utcoffset(self, dt):
        if isinstance(dt, datetime) or dt is None:
            return self._offset
        raise TypeError("utcoffset() argument must be a datetime instance"
                        " or None")

    def tzname(self, dt):
        if isinstance(dt, datetime) or dt is None:
            if self._name is None:
                return self._name_from_offset(self._offset)
            return self._name
        raise TypeError("tzname() argument must be a datetime instance"
                        " or None")

    def dst(self, dt):
        if isinstance(dt, datetime) or dt is None:
            return None
        raise TypeError("dst() argument must be a datetime instance"
                        " or None")

    def fromutc(self, dt):
        if isinstance(dt, datetime):
            if dt.tzinfo is not self:
                raise ValueError("fromutc: dt.tzinfo "
                                 "is not self")
            return dt + self._offset
        raise TypeError("fromutc() argument must be a datetime instance"
                        " or None")

    _maxoffset = timedelta(hours=23, minutes=59)
    _minoffset = -_maxoffset

    @staticmethod
    def _name_from_offset(delta):
        if delta < timedelta(0):
            sign = '-'
            delta = -delta
        else:
            sign = '+'
        hours, rest = divmod(delta, timedelta(hours=1))
        minutes = rest // timedelta(minutes=1)
        return 'UTC{}{:02d}:{:02d}'.format(sign, hours, minutes)

timezone.utc = timezone._create(timedelta(0))
timezone.min = timezone._create(timezone._minoffset)
timezone.max = timezone._create(timezone._maxoffset)


EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)


def datetime_to_timestamp(t):
    from time import mktime
    if t.tzinfo is None:
        return mktime((t.year, t.month, t.day,
                       t.hour, t.minute, t.second,
                       -1, -1, -1)) + t.microsecond / 1e6
    else:
        return (t - EPOCH).total_seconds()
